"""
Base class for all Hoppr plug-ins
"""
from __future__ import annotations

import functools
import subprocess
import time
import traceback

from abc import ABC, abstractmethod
from os import PathLike
from typing import Any, Callable, Dict, List, Optional

from packageurl import PackageURL

from hoppr import plugin_utils
from hoppr.mem_logger import MemoryLogger
from hoppr.models import HopprContext
from hoppr.models.transfer import ComponentCoverage
from hoppr.models.types import BomAccess
from hoppr.result import Result
from hoppr.utils import obscure_passwords


def _get_component(*args, **kwargs) -> Optional[Any]:
    # TODO: This will only utilize the first Component found in 1) args, 2) kwargs  # pylint: disable=fixme
    # and assumes all plugins operate this way
    comp = None
    for arg in args:
        if type(arg).__name__ == "Component":
            comp = arg
            break

    if comp is None:
        for value in kwargs.values():
            if type(value).__name__ == "Component":
                comp = value
                break

    return comp


def hoppr_process(func: Callable) -> Callable:
    """
    Decorator to handle generic bookkeeping for hoppr plug-ins
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs) -> Result:
        self._start_time = time.time()  # pylint: disable=protected-access
        self.create_logger()

        comp = _get_component(*args, **kwargs)

        arg_string = ""
        if comp is not None:
            arg_string = f"(purl: {comp.purl})"

        self.get_logger().debug(f"Starting {self.__class__.__name__}.{func.__name__} {arg_string}")

        result = None
        if comp is not None:
            if comp.purl is None:
                result = Result.fail("No purl supplied for component")

            else:
                purl_type = PackageURL.from_string(comp.purl).type

                if not self.supports_purl_type(purl_type):
                    return Result.skip(f"Class {self.__class__.__name__} does not support purl type {purl_type}")

                component_header = f"{comp.name}" + (f"@{comp.version}" if comp.version is not None else "")

                self.get_logger().info(
                    msg=f"{'-' * 4} Component: {component_header} {'-' * 50}",
                )

        if result is None:

            # Only check for missing commands if func has been overridden
            if func.__module__ != HopprPlugin.__module__:
                command_result = plugin_utils.check_for_missing_commands(self.required_commands)
                if command_result.is_fail():
                    self.get_logger().error(command_result.message)
                    return command_result

            try:
                result = func(self, *args, **kwargs)
            except Exception as error:  # pylint: disable=broad-except
                self.get_logger().error(
                    f"Unexpected exception running {self.__class__.__name__}.{func.__name__}: {error}",
                )
                self.get_logger().error(traceback.format_exc())
                result = Result.fail(f"Unexpected Exception: {error}")

        self.get_logger().debug(f"Completed {self.__class__.__name__}.{func.__name__}")

        duration = time.time() - self._start_time  # pylint: disable=protected-access
        self.get_logger().debug(f"Process duration {duration:3f} seconds")
        self.get_logger().info(f"Result: '{result}'")

        if result.is_skip():
            self.get_logger().clear_targets()

        self.close_logger()

        return result

    return wrapper


def hoppr_rerunner(method: Callable) -> Callable:
    """
    Runs a method (assumed to return a Result object) a number of times,
    or until the result type is not RETRY
    """

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs) -> Result:

        log = self.get_logger()

        result = Result.fail(message="Context max_attempts must be a positive integer.")

        for attempt in range(self.context.max_attempts):
            log.info(msg=f"Processing component [attempt {attempt + 1} of {self.context.max_attempts}]", indent_level=1)

            result = method(self, *args, **kwargs)

            if not isinstance(result, Result):
                msg = f"Method {method.__name__} returned {type(result).__name__} in rerunner. Result object required"

                log.error(msg=msg, indent_level=1)
                return Result.fail(message=msg)

            if not result.is_retry():
                return result

            if attempt < self.context.max_attempts - 1:
                log.warning(
                    msg=(f"Method {method.__name__} will be retried in {self.context.retry_wait_seconds} seconds"),
                    indent_level=1,
                )

                log.warning(msg=f"Result message for attempt {attempt + 1}: {result.message}", indent_level=1)

                time.sleep(self.context.retry_wait_seconds)

        log.error(msg=f"Method {method.__name__} failed after {self.context.max_attempts} attempts")
        log.error(msg=f"Result message for final attempt: {result.message}")

        return Result.fail(f"Failure after {self.context.max_attempts} attempts, final message {result.message}")

    return wrapper


def hoppr_ignore_excluded(func: Callable) -> Callable:
    """
    Decorator to ignore components which have a scope of `excluded`
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs) -> Result:

        comp = _get_component(*args, **kwargs)
        if comp is not None and str(comp.scope) in ('excluded', 'Scope.excluded'):
            return Result.excluded()
        return func(self, *args, **kwargs)

    return wrapper


class HopprPlugin(ABC):
    """
    Base class for all Hoppr plug-ins

    Note that this class is not thread safe
    """

    required_commands: List[str] = []
    supported_purl_types: List[str] = []

    """
    Product: the result of carrying out a step. Products are usually persistent (e.g,.
    files), and are often meant to be used as materials on subsequent steps. Products are
    recorded as part of link metadata.

    This is a list of strings where each element is a relative path file or wildcard.

    See in-toto specificaion:
    - https://github.com/in-toto/docs/blob/master/in-toto-spec.md
    """
    products: List[str] = []

    default_component_coverage = ComponentCoverage.OPTIONAL
    bom_access = BomAccess.NO_ACCESS
    process_timeout = 60

    def __init__(
        self,
        context: HopprContext,
        config: Optional[Dict] = None,
    ) -> None:
        self._logger: MemoryLogger
        self._start_time: float = 0.0
        self.context = context
        self.config = config

        if self.config is not None:
            if "process_timeout" in self.config:
                self.process_timeout = self.config["process_timeout"]

    @abstractmethod
    def get_version(self) -> str:
        """
        Returns the version of this plug-in
        """

    @hoppr_process
    @hoppr_rerunner
    def pre_stage_process(self) -> Result:
        """
        Process to be run before other processing within a stage for this plug-in
        """

        return Result.skip("pre_stage_process not defined.")

    @hoppr_process
    @hoppr_rerunner
    def process_component(self, comp: Any) -> Result:  # pylint: disable=unused-argument
        """
        Process a single component through this plug-in
        """

        return Result.skip("process_component not defined.")

    @hoppr_process
    @hoppr_rerunner
    def post_stage_process(self):
        """
        Finalize processing for this plug-in
        """
        return Result.skip("post_stage_process not defined.")

    def supports_purl_type(self, purl_type: str) -> bool:
        """
        Indicates whether or not this particular plug-in supports components of the specified
        purl type.

        If no supported purl types are defined, the plug-in supports all purl types.
        """
        return purl_type in self.supported_purl_types or not self.supported_purl_types

    @classmethod
    def get_attestation_products(cls, config: dict | None = None) -> list[str]:  # pylint: disable=unused-argument
        """
        Return a list of attestation products for this class
        """

        return cls.products

    def create_logger(self) -> None:
        """
        Creates a logger with a MemoryHandler, which flushes to stderr when closed
        """
        self._logger = MemoryLogger(
            file_name=str(self.context.logfile_location),
            lock=self.context.logfile_lock,  # type: ignore[arg-type]
            log_level=self.context.log_level,
        )

    def get_logger(self) -> MemoryLogger:
        """
        Returns the logger to be used for the current process
        """
        return self._logger

    def close_logger(self) -> None:
        """
        Close (and flush) all handlers for this plug-in's logger
        """
        self._logger.close()

    def run_command(
        self, command: list[str], password_list: list[str] | None = None, cwd: str | PathLike[str] | None = None
    ) -> subprocess.CompletedProcess[bytes]:
        """
        Run a command and log any errors
        """
        obscured_command = obscure_passwords(command, password_list)
        self.get_logger().debug(msg=f"Running command: '{obscured_command}'", indent_level=2)

        try:
            result = subprocess.run(
                command, check=False, shell=False, capture_output=True, cwd=cwd, timeout=self.process_timeout
            )
        except subprocess.TimeoutExpired as timeout_expired:
            self.get_logger().error(
                f"Command \"{command[0]}\" timed out after {timeout_expired.timeout} seconds", indent_level=2
            )
            result = subprocess.CompletedProcess(
                command, returncode=124, stdout=timeout_expired.stdout or b'', stderr=timeout_expired.stderr or b''
            )

        if result.returncode != 0:
            # Prefix each line of the command's stdout with 4 spaces
            stdout_indented = "\n".join(["    " + line for line in result.stdout.decode("utf-8").split("\n")])

            self.get_logger().debug(msg=f"{command[0]} command stdout content:\n{stdout_indented}", indent_level=2)

            # Prefix each line of the command's stderr with 4 spaces
            stderr_indented = "\n".join(["    " + line for line in result.stderr.decode("utf-8").split("\n")])

            self.get_logger().error(msg=f"{command[0]} command failed with error:\n{stderr_indented}", indent_level=2)

        return result



"""
Module including functions and classes for working with the in-toto spec.

This module includes a function for creating in-toto layout files, and
wrapping other processing with "start" and "stop" functions.
"""
from __future__ import annotations

from pathlib import Path

from in_toto.models.layout import Layout, Step  # type: ignore
from in_toto.models.metadata import Metablock  # type: ignore
from in_toto.runlib import in_toto_record_start, in_toto_record_stop  # type: ignore
from securesystemslib import interface  # type: ignore

from hoppr.models.transfer import Transfer
from hoppr.utils import plugin_class


def _get_products(transfer: Transfer) -> tuple[dict[str, list[str]], list[str]]:
    """
    Function to determine what products are produced for each "hoppr stage"/"in-toto step".

    This is done by looking at all plugins in a stage, and create a list of "products". There
    are two additional "products" included in each step/stage which are a consolidated bom
    and a delivered bom.
    """
    products = {}
    stages = []

    stages.append("_collect_metadata")
    products["_collect_metadata"] = ["generic/_metadata_/*"]

    for stage_ref in transfer.stages:

        stages.append(stage_ref.name)

        products[stage_ref.name] = []
        for plugin_ref in stage_ref.plugins:
            plugin_cls = plugin_class(plugin_ref.name)
            products[stage_ref.name] += plugin_cls.get_attestation_products(plugin_ref.config)

        products[stage_ref.name] += [
            "generic/_metadata_/_delivered_bom.json",
            f"generic/_metadata_/_intermediate_{stage_ref.name}_delivered_bom.json",
        ]

    stages.append("_finalize")
    products["_finalize"] = ["generic/_metadata_/_delivered_bom.json"]

    return (products, stages)


def generate_in_toto_layout(  # pylint: disable=too-many-locals
    transfer: Transfer,
    project_owner_key_path: Path,
    functionary_key_path: Path,
    project_owner_key_password: str | None = None,
    relative_expiration: int = 6,
) -> None:
    """
    Function to create an in-toto layout file based on a transfer object.

    The layout file will have "steps" where each "step" is equal to a hoppr stage. All files changed in a stage
    must be referenced in the "step" to have a verifiable layout. All plugins have a `products` property which
    lists the expected files it will change. The "step" products is a list of all of these hoppr plugin `products`
    of the stage.

    Additionally there are two additional "steps":
    1. _collect_metadata - wraps the `_collect_metadata()` called before hoppr processes stages
    1. _finalize - wraps the `_collect_bom()`  called after hoppr processes stages
    """

    layout = Layout()
    layout.set_relative_expiration(months=relative_expiration)

    project_owner_key = interface.import_privatekey_from_file(
        str(project_owner_key_path), password=project_owner_key_password
    )
    functionary_pubkey = layout.add_functionary_key_from_path(str(functionary_key_path) + ".pub")

    (products, stages) = _get_products(transfer)

    for i, stage in enumerate(stages):
        step = Step(name=stage, pubkeys=[functionary_pubkey["keyid"]])

        for j in range(0, i):
            step.add_material_rule_from_string(f"MATCH * WITH PRODUCTS FROM {stages[j]}")

        for product in products[stage]:
            step.add_product_rule_from_string(f"ALLOW {product}")

        step.add_material_rule_from_string("DISALLOW *")
        step.add_product_rule_from_string("DISALLOW *")

        layout.steps.append(step)

    metablock = Metablock(signed=layout)
    metablock.sign(project_owner_key)
    metablock.dump("in-toto.layout")


class HopprInTotoLinks:
    """
    Class for managing state of in-toto "start" and "stop" function calls.

    The intent is to simplify the creation of in-toto link (aka attestation) files.
    These methods are used to "wrap" hoppr processing capabilities to track changes
    to the temporary hoppr storage location.

    This wrapping of hoppr processing capabilities should be on stages (not plugins)
    to allow parallel processing of plugins.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        create_attestations: bool,
        transfer: Transfer,
        functionary_key_path: Path | None = None,
        functionary_key_password: str | None = None,
        metadata_directory: str | None = None,
    ) -> None:
        self.create_attestations = create_attestations

        if self.create_attestations:
            self.products, _ = _get_products(transfer)
            self.functionary_key = interface.import_privatekey_from_file(
                filepath=str(functionary_key_path), password=functionary_key_password
            )
            self.collection_root = ""
            self.metadata_directory = metadata_directory

    def set_collection_root(self, collection_root):
        """
        Setter method for the collection root location.

        Since the collection root is a temporary directory, its created after
        the hoppr_in_toto_links object is created.
        """
        self.collection_root = collection_root

    def record_stage_start(self, stage_name: str):
        """
        Method called just before a stage is processed.

        This will call the in-toto framework and record all files within the hoppr temporary
        processing folder.
        """
        if self.create_attestations:
            in_toto_record_start(
                step_name=stage_name,
                material_list=["./"],
                signing_key=self.functionary_key,
                base_path=self.collection_root,
                record_environment=True,
            )

    def record_stage_stop(self, stage_name: str):
        """
        Method called just after a stage is processed.

        This will call the in-toto framework and record all products files produced based on the
        expected products of the plugins running in the stage.
        """

        if self.create_attestations:
            # Use pathlib to find files using wildcard in names
            posix_path_list: list[Path] = []
            for product in self.products[stage_name]:
                posix_path_list += Path(self.collection_root).glob(product)

            # After finding paths, convert them to strings that are a relative path from
            # the collect root directory.
            product_list = [str(posix_path.relative_to(self.collection_root)) for posix_path in posix_path_list]

            in_toto_record_stop(
                step_name=stage_name,
                product_list=product_list,
                signing_key=self.functionary_key,
                base_path=self.collection_root,
                metadata_directory=self.metadata_directory,
            )



"""
Plugin to create tar file bundle
"""
import os
import re
import tarfile
import time

from hoppr import __version__
from hoppr.base_plugins.hoppr import HopprPlugin, hoppr_process
from hoppr.result import Result


class TarBundlePlugin(HopprPlugin):
    """
    Plugin to create tar file bundle

    This plug-in supports the following config values:
       - compression: Specifies the compression to be applied to the output tar file.  Defaults
                      to "gz".  Recognized values are "none", "gzip", "gz", "bzip2", "bz",
                      "lmza", "xz"

        - tarfile_name: Name of the tar file to be created.  Defaults to "./bundle.tar.<comp>",
                        where <comp> is the short form of the compression used.  For "none"
                        compression, default is "./bundle.tar"
    """

    def get_version(self) -> str:
        return __version__

    @hoppr_process
    def post_stage_process(self):
        """
        Tar-up the context.collect_root_dir directory
        """

        compression = "gz"
        if self.config is not None:
            compression = self.config.get("compression", "gz").lower()

        if compression == "none":
            compression = ""
            default_filename = os.sep.join([".", "bundle.tar"])
        elif compression in ["gzip", "gz"]:
            compression = "gz"
            default_filename = os.sep.join([".", "bundle.tar.gz"])
        elif compression in ["bzip2", "bz"]:
            compression = "bz2"
            default_filename = "./bundle.tar.bz"
        elif compression in ["lzma", "xz"]:
            compression = "xz"
            default_filename = os.sep.join([".", "bundle.tar.xz"])
        else:
            return Result.fail(f"Unrecognized compression in config: {compression}")

        root_dir = self.context.collect_root_dir

        tarfile_name = default_filename
        if self.config is not None:
            tarfile_name = self.config.get("tarfile_name", default_filename)
            tarfile_name = os.path.expanduser(tarfile_name)

        if os.path.exists(tarfile_name):
            timestr = time.strftime("%Y%m%d-%H%M%S")
            matches = re.search(r"^(.*?)(\.tar(\..*)?)?$", tarfile_name)

            if matches:
                tarfile_name = f"{matches.group(1)}-{timestr}{matches.group(2) if matches.group(2) is not None else ''}"

        self.get_logger().info(f"Bundling collected artifacts into {tarfile_name}")

        try:
            with tarfile.open(tarfile_name, f"x:{compression}") as tar:
                tar.add(root_dir, ".")
        except tarfile.ReadError as err:
            return Result.fail(f"Unable to create tarfile {tarfile_name}, {err}")
        except FileNotFoundError:
            return Result.fail(f"File {tarfile_name}: Directory not found.")
        except PermissionError:
            return Result.fail(f"File {tarfile_name}: Permission denied.")

        return Result.success()



"""
Report Generating Plugin
"""

import json
import shutil

from pathlib import Path
from typing import Any, Optional
from uuid import UUID

import typer

from jinja2 import Environment, FileSystemLoader

from hoppr import __version__
from hoppr.base_plugins.hoppr import HopprPlugin
from hoppr.models.manifest import Component
from hoppr.result import ResultStatus


class Report:  # pylint: disable=too-few-public-methods
    """
    A report of a plugin's execution stage result.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        unique_id: UUID,
        plugin: str,
        stage: str,
        result: ResultStatus,
        details: Any,
        method: str,
        component: Optional[Component] = None,
    ):
        self.unique_id = unique_id
        self.plugin = plugin
        self.stage = stage
        self.result = result
        self.details = details
        self.method = method
        self.component = component


class ReportGenerator(HopprPlugin):
    """
    A class for generating a report based on a list of `Report` objects.
    """

    report_gen_list: list[Report] = []

    def get_version(self) -> str:
        return __version__

    def generate_report(self) -> None:  # pylint: disable=too-many-locals
        """
        Generate a report based on the given list of `Report` objects.
        """

        # Count the number of reports and errors
        report_count = len(self.report_gen_list)
        error_count = len([report for report in self.report_gen_list if report.result.numerator == ResultStatus.FAIL])

        # Sort reports
        reports_with_components = [report for report in self.report_gen_list if report.component is not None]
        sorted_reports = sorted(reports_with_components, key=lambda report: (-report.result.numerator, report.plugin))

        # Organize reports by plugin and stage
        reports_by_plugin = self._get_reports_by("plugin", sorted_reports)
        reports_by_stage = self._get_reports_by("stage", sorted_reports)

        # Generate report HTML file
        environment = Environment(loader=FileSystemLoader(Path(__file__).parent / "templates"))
        index_template = environment.get_template("index.jinja2")

        report_dir = Path.cwd() / "report"
        report_dir.mkdir(exist_ok=True)

        results_file_name = report_dir / "index.html"

        for dir_name in ("assets", "scripts", "styles"):
            source_dir = Path(__file__).parent / dir_name
            dest_dir = report_dir / dir_name
            shutil.copytree(src=source_dir, dst=dest_dir, dirs_exist_ok=True)

        context = {
            "report_count": report_count,
            "error_count": error_count,
            "reports_by_stage": reports_by_stage,
            "reports_by_plugin": reports_by_plugin,
            "ResultStatus": ResultStatus,
            "get_type": type,
            "get_length": len,
            "json_loads": json.loads,
            "get_successful_report_total": self._get_successful_report_total,
            "get_overall_status_button_class": self._get_overall_status_button_class,
        }

        results_file_name.write_text(index_template.render(context), encoding="utf-8")
        typer.echo(f"HTML report generated: {results_file_name}")

    def _get_successful_report_total(self, reports: list[Report]) -> int:
        """
        Return the number of successful reports for a given list of reports.
        """

        successful_reports = [report for report in reports if report.result.numerator == ResultStatus.SUCCESS]
        return len(successful_reports)

    def _get_overall_status_button_class(self, reports: list[Report]) -> str:
        """
        Determine the overall status of a list of reports and returns the appropriate css class.
        """

        num_successful_reports = sum(report.result.numerator == ResultStatus.SUCCESS for report in reports)

        if num_successful_reports == 0:
            return "btn-danger"

        if num_successful_reports == len(reports):
            return "btn-success"

        return "btn-warning"

    def _get_reports_by(self, sort_by: str, reports: list[Report]) -> dict:
        """
        Organize report data by plugin or stage.
        """

        sorted_reports: dict = {}

        for report in reports:
            if report.__dict__[sort_by] in sorted_reports:
                sorted_reports[report.__dict__[sort_by]].append(report)
            else:
                sorted_reports[report.__dict__[sort_by]] = [report]

        return sorted_reports



"""
Base class for all collector plugins
"""
from __future__ import annotations

import json
import socket

from abc import abstractmethod
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Union, final
from urllib.parse import urlparse

from packageurl import PackageURL  # type: ignore[import]

from hoppr import __version__, constants, plugin_utils
from hoppr.base_plugins.hoppr import HopprPlugin, hoppr_ignore_excluded, hoppr_process, hoppr_rerunner
from hoppr.models.credentials import CredentialRequiredService, Credentials
from hoppr.models.manifest import Component, Property, SearchSequence
from hoppr.models.transfer import ComponentCoverage
from hoppr.models.types import BomAccess, PurlType
from hoppr.result import Result
from hoppr.utils import dedup_list, remove_empty


class BaseCollectorPlugin(HopprPlugin):
    """
    Base class for collector plugins
    """

    default_component_coverage = ComponentCoverage.EXACTLY_ONCE
    bom_access = BomAccess.COMPONENT_ACCESS
    system_repositories: list[str] = []

    @staticmethod
    def _get_repo_search_list(comp: Component) -> list[str]:
        repo_list = []
        for prop in comp.properties or []:
            if prop.name == constants.BomProps.COMPONENT_SEARCH_SEQUENCE:
                search_sequence_str = json.loads(prop.value or '{"version": "v1", "repositories": []}')
                search_sequence = SearchSequence.parse_obj(search_sequence_str)
                repo_list.extend(search_sequence.repositories)

        return repo_list

    def _get_repos(self, comp: Component) -> list[str]:
        """
        Returns all repos listed in all BOM_PROPS_COMPONENT_SEARCH_SEQUENCE properties for this component
        """

        purl_url = PackageURL.from_string(comp.purl).qualifiers.get("repository_url")

        if purl_url is not None and not self.context.strict_repos:
            return [purl_url]

        repo_list = self._get_repo_search_list(comp)

        if purl_url is not None:
            parsed_purl_url = self._parse_url_for_comparison(purl_url)
            for repo in repo_list:
                parsed_repo_url = self._parse_url_for_comparison(repo)
                if (
                    parsed_repo_url[0] == parsed_purl_url[0]
                    and (not parsed_purl_url[1] or parsed_repo_url[1] == parsed_purl_url[1])
                    and parsed_repo_url[2].startswith(parsed_purl_url[2])
                ):
                    return [purl_url]

            return []

        if not self.context.strict_repos:
            repo_list.extend(self.system_repositories)

        return dedup_list(repo_list)

    def empty_repo_list_reason(self, comp: Component) -> str:
        """
        _get_repos may return an empty list, this method returns a string
        giving a reasonable explanation as to why.
        """
        purl = PackageURL.from_string(comp.purl)

        if not self.context.strict_repos:
            return f"No repositories specified for purl type {purl.type}, and no repository_url specified in purl"

        if len(self._get_repo_search_list(comp)) == 0:
            return (
                f"No repositories specified in manifest for purl type {purl.type}. "
                "Add repositories or use --no-strict command line option"
            )

        if purl.qualifiers.get("repository_url") is not None:
            return (
                f"Specified repository_url ({purl.qualifiers.get('repository_url')}) is not in "
                "manifest repository list.  Add it to the manifest, or use --no-strict command "
                "line option"
            )

        return "Reason for empty repo list undetermined"

    def directory_for(
        self,
        purl_type: Union[str, PurlType],
        repo_url: str,
        subdir: Union[str, None] = None,
    ) -> Path:
        """
        Identify the directory into which the artifact should be copied
        """

        repo_dir = plugin_utils.dir_name_from_repo_url(repo_url)
        directory = Path(self.context.collect_root_dir, str(purl_type), repo_dir)

        if subdir is not None:
            directory = Path(directory, subdir)

        directory.mkdir(parents=True, exist_ok=True)

        return directory

    @staticmethod
    def _parse_url_for_comparison(url: str) -> tuple[str, str, str]:
        """
        Resulting tuple is (host, port, path)
        """
        if url.startswith("file:") or "//" in url:
            initial_parse = urlparse(url)
        else:
            initial_parse = urlparse("//" + url)

        host_data = initial_parse.netloc.split(":")

        host = host_data[0]

        if len(host_data) > 1:
            port = host_data[1]
        elif initial_parse.scheme == "":
            port = ""
        else:
            port = str(socket.getservbyname(initial_parse.scheme))

        path = initial_parse.path.rstrip("/")

        return (host, port, path)

    @staticmethod
    def check_purl_specified_url(purl: PackageURL, repo_url: str) -> Result:
        """
        Test if a repository_url specified as a purl qualifier is a mis-match with a given repo_url

        If no repository_url is specified, repo_url is fine, no-mismatch, return False

        Otherwise, considered a match if repo_url starts with the purl url (after trimming trailing
        "/"s and URL scheme, if present)
        """
        purl_url = purl.qualifiers.get("repository_url")
        if purl_url is None:
            return Result.success()

        parsed_purl_url = BaseCollectorPlugin._parse_url_for_comparison(purl_url)
        parsed_repo_url = BaseCollectorPlugin._parse_url_for_comparison(repo_url)

        if (
            parsed_repo_url[0] == parsed_purl_url[0]  # Compare host
            and (not parsed_purl_url[1] or parsed_repo_url[1] == parsed_purl_url[1])  # Compare port if present
            and parsed_repo_url[2].startswith(parsed_purl_url[2])  # Compare initial path
        ):
            return Result.success()

        return Result.fail(f"Purl-specified repository url ({purl_url}) does not match current repo ({repo_url}).")

    def set_collection_params(self, comp: Component, repository: str, directory: Union[Path, str]) -> None:
        """
        Set collection parameters on sbom component
        """

        rel_dir = str(Path(directory).relative_to(self.context.collect_root_dir))

        collect_props: dict[str, str] = {
            constants.BomProps.COLLECTION_REPOSITORY: repository,
            constants.BomProps.COLLECTION_DIRECTORY: rel_dir,
            constants.BomProps.COLLECTION_PLUGIN: f"{type(self).__name__}:{self.get_version()}",
            constants.BomProps.COLLECTION_TIMETAG: str(datetime.now(timezone.utc)),
        }

        if comp.properties is None:
            comp.properties = []

        for prop in comp.properties:
            if prop.name in collect_props:
                prop.value = collect_props.pop(prop.name)

        for key, value in collect_props.items():
            if value is not None:
                comp.properties.append(Property(name=key, value=value))

    def _check_collection_params(self, result: Result) -> Result:
        """
        All successful collections should set the collection parameters.
        This method makes sure that they do.
        """
        if not result.is_success():
            return result

        if not isinstance(result.return_obj, Component):
            return result.fail(
                f"Collector class {type(self).__name__} process_component method returned a successful result "
                "without updating the BOM component."
            )

        missing_props = []
        for req_prop in [
            constants.BomProps.COLLECTION_REPOSITORY,
            constants.BomProps.COLLECTION_DIRECTORY,
            constants.BomProps.COLLECTION_PLUGIN,
            constants.BomProps.COLLECTION_TIMETAG,
        ]:

            if req_prop.value not in [prop.name for prop in result.return_obj.properties or []]:
                missing_props.append(req_prop.value)

        if missing_props:
            return result.fail(
                f"Collector class {type(self).__name__} process_component method returned a successful result "
                f"without updating the following component properties: {', '.join(missing_props)}"
            )

        return result


class SerialCollectorPlugin(BaseCollectorPlugin):
    """
    Base class for multi-process collector plugins
    """

    @abstractmethod
    @hoppr_rerunner
    def collect(self, comp: Any, repo_url: str, creds: CredentialRequiredService | None = None):
        """
        This method should attempt to collect a single component from the specified URL
        """

    @final
    @hoppr_process
    @hoppr_ignore_excluded
    def process_component(self, comp: Component) -> Result:
        """
        Copy a component to the local collection directory structure

        A CollectorPlugin will never return a RETRY result, but handles the retry logic internally.
        """

        logger = self.get_logger()

        repo_list = self._get_repos(comp)
        if len(repo_list) == 0:
            return Result.fail(self.empty_repo_list_reason(comp))

        for repo_url in repo_list:
            logger.info(f"Repository: {repo_url}")

            repo_creds = Credentials.find(repo_url)
            result = self.collect(comp, repo_url, repo_creds)

            if result.is_success():
                break  ### We found it, no need to try any more repositories

        return self._check_collection_params(result)

    @hoppr_process
    def post_stage_process(self):
        for purl_type in self.supported_purl_types:
            directory = Path(self.context.collect_root_dir, purl_type)

            if directory.is_dir():
                remove_empty(directory)

        return Result.success()


class BatchCollectorPlugin(BaseCollectorPlugin):
    """
    Base class for single-process collector plugins
    """

    config_file: Path

    @abstractmethod
    @hoppr_rerunner
    def collect(self, comp: Component):
        """
        This method should attempt to collect all components
        from all manifest repositories or registries that were
        previously configured in the pre stage process.

        Use of a single batch operation (i.e. dynamically constructed
        shell command) is encouraged if supported by the underlying
        collection tool(s).
        """

    @final
    @hoppr_process
    @hoppr_ignore_excluded
    def process_component(self, comp: Component) -> Result:
        """
        Copy a component to the local collection directory structure

        A CollectorPlugin will never return a RETRY result, but handles the retry logic internally.
        """
        logger = self.get_logger()
        logger.info(f"Processing component {comp.purl}")

        return self._check_collection_params(self.collect(comp))



"""
Base model for Hoppr config files
"""
from __future__ import annotations

import logging

from copy import deepcopy
from multiprocessing.managers import BaseProxy
from pathlib import Path
from threading import _RLock as RLock
from typing import TYPE_CHECKING, ClassVar

from pydantic import Field, root_validator

from hoppr.models.base import HopprBaseModel
from hoppr.models.credentials import CredentialRequiredService, Credentials, CredentialsFile
from hoppr.models.manifest import Manifest, ManifestFile, Repositories, Sbom
from hoppr.models.transfer import StageRef, Transfer, TransferFile

__all__ = [
    "Credentials",
    "CredentialsFile",
    "HopprBaseModel",
    "Manifest",
    "ManifestFile",
    "Transfer",
    "TransferFile",
]

if TYPE_CHECKING:
    from pydantic.typing import DictStrAny


class HopprSchemaModel(HopprBaseModel):
    """
    Consolidated Hoppr config file schema definition
    """

    __root__: CredentialsFile | ManifestFile | TransferFile = Field(..., discriminator="kind")


class HopprContext(HopprBaseModel):
    """
    Consolidated data model for Hoppr runtime context
    """

    collect_root_dir: Path
    consolidated_sbom: Sbom
    credential_required_services: list[CredentialRequiredService] | None
    delivered_sbom: Sbom
    log_level: int = logging.INFO
    logfile_location: Path = Path("hoppr.log")
    logfile_lock: BaseProxy | RLock  # Union type to fix pydantic validation of multiprocessing library
    max_attempts: int = 3
    max_processes: int
    previous_delivery: Path | str | None = None
    repositories: Repositories
    retry_wait_seconds: float = 5
    sboms: list[Sbom]
    stages: list[StageRef]
    strict_repos: bool = True



"""
Plugin to create oras oci bundle
"""
import re

from datetime import datetime
from pathlib import Path
from typing import Any, List, Tuple

import oras.client
import oras.defaults
import oras.oci
import oras.provider
import oras.utils

from oras.container import Container

from hoppr import __version__
from hoppr.base_plugins.hoppr import HopprPlugin, hoppr_process
from hoppr.core_plugins.oras_registry import Registry
from hoppr.exceptions import HopprCredentialsError, HopprPluginError
from hoppr.models import HopprContext as Context
from hoppr.models.credentials import Credentials
from hoppr.result import Result


class OrasBundlePlugin(HopprPlugin):
    """
    Plugin to create an Oras file upload

    This plug-in supports the following config values:
        - oras_artifact_name: Name of oras artifact.
        - oras_artifact_version: Version of artifact.
        - oras_registry: Registry to push Oras Artifact.
    """

    def get_version(self) -> str:
        return __version__

    def __init__(self, context: Context, config: Any = None) -> None:
        super().__init__(context=context, config=config)
        self._results: List[Tuple[Path, Result]] = []
        self.failures = 0
        self.retries = 0
        self.finalize_failures = 0

    @hoppr_process
    def post_stage_process(self):
        """
        Bundle up the context.collect_root_dir directory and push to registry
        """

        root_dir = self.context.collect_root_dir

        if self.config is not None:
            oras_artifact_name = self.config.get("oras_artifact_name")
            if oras_artifact_name is None:
                raise HopprPluginError("Failed to collect oras artifact name from config.")
            oras_artifact_version = self.config.get("oras_artifact_version")
            if oras_artifact_version is None:
                raise HopprPluginError("Failed to collect oras artifact version from config.")
            oras_registry = self.config.get("oras_registry")
            if oras_registry is None:
                raise HopprPluginError("Failed to collect oras registry from config.")
            if oras_registry.startswith('http'):
                raise HopprPluginError("Oras Registry name should just be a hostname and not contain a protocol scheme")
        else:
            self.get_logger().info("Oras config not correct")
            return Result.fail()

        self.get_logger().info(
            "Bundling collected artifacts into oras artifact " f"{oras_artifact_name}:{oras_artifact_version}"
        )
        self.get_logger().flush()

        if ':' in oras_registry:
            cred = Credentials.find(oras_registry.split(':', maxsplit=1)[0])
        else:
            cred = Credentials.find(oras_registry)
        if cred is None:
            raise HopprCredentialsError("Credentials must not be empty for Oras Bundle Plugin")

        # Setup Oras Client
        client = self.get_oras_client(username=cred.username, password=cred.password.get_secret_value())

        # Gather list of files to upload
        file_list = [str(file) for file in root_dir.rglob("*") if file.is_file()]
        archives = self.get_files_from_root_dir(file_list=file_list, root_dir=root_dir)
        archives = self.verify_contents(self.context.delivered_sbom.components, archives)

        # Push should be relative to cache context
        with oras.utils.workdir(root_dir):
            uri = f'{oras_artifact_name}:{oras_artifact_version}'
            container = Container(name=uri, registry=oras_registry)
            client.push(container, archives, logger=self.get_logger())
        self.get_logger().info(f"Uploaded: {uri}")
        return Result.success()

    def get_oras_client(self, username: str, password: str):
        """
        Consistent method to get an oras client

        Args:
            username (str): username for registry
            password (str): password for registry

        Returns:
            hoppr.core_plugins.oras_registry.Registry: Override of oras registry object
        """
        reg = Registry()
        if username and password:
            self.get_logger().info("Found username and password for basic auth")
            reg.set_basic_auth(username, password)
        else:
            raise HopprPluginError("Username and Password not set in the credentials.yml.")
        return reg

    def get_files_from_root_dir(self, file_list: list, root_dir: Path) -> list:
        """
        Helper function to get file array list and build archives

        Args:
            root_dir (Path): _description_

        Returns:
            list: A list of dictionaries
        """
        # Create lookup of archives - relative path and mediatype
        archives = []
        now = datetime.now()

        for filename in file_list:
            media_type = "application/vnd.oci.image.layer.v1.tar"
            size = Path(filename).stat().st_size  # bytes
            annotations = {"creationTime": str(now), "size": str(size)}
            if str(root_dir / "generic" / "_metadata_" / "_delivered_bom.json") == filename:
                media_type = "application/vnd.cyclonedx"
            elif str(root_dir / "generic" / "_metadata_" / "_consolidated_bom.json") == filename:
                media_type = "application/vnd.cyclonedx"
                media_type = 'application/vnd.cyclonedx'
            archives.append(
                {
                    "path": filename,
                    "title": filename.replace(f"{root_dir}/", "", -1),
                    "media_type": media_type,
                    "annotations": annotations,
                }
            )
        return archives

    def verify_contents(self, components: list, archives: list) -> list:
        """
        Verify that components are upgraded

        Args:
            components (list): Cyclonedx Components
            archives (list): List of dictionaries

        Raises:
            HopprPluginError: Hoppr plugin error
        """
        total_components = len(components)
        self.get_logger().info(f"Total components: {total_components}")
        self.get_logger().info(f"Total archives: {len(archives)}")
        self.get_logger().flush()
        for component in components:
            pattern = f"{component.name}[-|_]{component.version}"
            scope = str(component.scope)
            for archive in archives:
                result = re.match(pattern, archive['path'])
                if result:
                    if 'excluded' in scope:
                        self.get_logger().info(f"Found scope {component.scope}, removing component from archive.")
                        self.get_logger().flush()
                        archives.remove(archive)
        self.get_logger().info("Components validated in bundle for upload.")
        self.get_logger().flush()
        return archives



"""
Transfer file data model
"""
from __future__ import annotations

import math
import re

from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, Literal, Pattern

from pydantic import ConstrainedStr, Field, validator

from hoppr.models.base import HopprBaseModel, HopprBaseSchemaModel

if TYPE_CHECKING:
    from pydantic.typing import DictStrAny
else:
    DictStrAny = dict[str, Any]


class StageName(ConstrainedStr):
    """
    Constrained string type for stage key name
    """

    regex: Pattern[str] = re.compile(pattern=r"^[A-Za-z_][\w\s]*$")
    min_length: int = 1


class Plugin(HopprBaseModel):
    """
    Plugin data model
    """

    name: str = Field(..., description="Name of plugin")
    config: DictStrAny | None = Field(None, description="Mapping of additional plugin configuration settings to values")


Plugin.update_forward_refs()


class ComponentCoverage(Enum):
    """
    Enumeration to indicate how often each component should be processed
    """

    OPTIONAL = (0, math.inf)
    EXACTLY_ONCE = (1, 1)
    AT_LEAST_ONCE = (1, math.inf)
    NO_MORE_THAN_ONCE = (0, 1)

    def __init__(self, min_allowed: int, max_allowed: int):
        self.min_value = min_allowed
        self.max_value = max_allowed

    def __str__(self) -> str:
        return str(self.name)

    def accepts_count(self, count: int) -> bool:
        """
        Identifies whether a specified count is acceptable for this coverage value
        """
        return self.min_value <= count <= self.max_value


class Stage(HopprBaseModel):
    """
    Stage data model
    """

    component_coverage: Literal["AT_LEAST_ONCE", "EXACTLY_ONCE", "NO_MORE_THAN_ONCE", "OPTIONAL"] | None = Field(
        default=None, exclude=True, description="Defines how often components should be processed"
    )
    plugins: list[Plugin] = Field(..., description="List of Hoppr plugins to load")


class StageRef(Stage):
    """
    StageRef data model
    """

    name: StageName


Stages = Annotated[dict[StageName, Stage], ...]


class TransferFile(HopprBaseSchemaModel):
    """
    Transfer file data model
    """

    kind: Literal["Transfer"]
    max_processes: int | None = Field(3, description="Max processes to create when running Hoppr application")
    stages: Stages = Field(..., description="Mapping of stage names to property definitions")


class Transfer(TransferFile):
    """
    Transfer data model
    """

    stages: list[StageRef]  # type: ignore[assignment]

    @validator("stages", allow_reuse=True, pre=True)
    @classmethod
    def validate_stages(cls, stages: DictStrAny) -> list[StageRef]:
        """
        Transform Stages into list of StageRef objects
        """
        stage_refs: list[StageRef] = []

        add_delta = True

        for stage_name, stage in stages.items():
            stage["name"] = stage_name
            stage_refs.append(StageRef.parse_obj(stage))

            for plugin in stage["plugins"]:
                if plugin["name"] == "hoppr.core_plugins.delta_sbom":
                    add_delta = False

        if add_delta:
            stage_refs.insert(
                0,
                StageRef(
                    name=StageName("_delta_sbom_"),
                    plugins=[
                        Plugin(
                            name="hoppr.core_plugins.delta_sbom",
                            config=None,
                        )
                    ],
                ),
            )

        return stage_refs

    @classmethod
    def load(cls, source: str | Path | DictStrAny) -> Transfer:
        """
        Load transfer file from local path or dict
        """
        match source:
            case dict():
                return cls.parse_obj(source)
            case str() | Path():
                return cls.parse_file(source)
            case _:
                raise TypeError("'source' argument must be one of: 'str', 'Path', 'dict[str, Any]'")



"""
Base model for Hoppr config files
"""
from __future__ import annotations

from typing import Literal

from pydantic import Extra, Field, validator
from pydantic_yaml import YamlModel


class Metadata(YamlModel):
    """
    Metadata data model
    """

    name: str
    version: str | int
    description: str


class HopprBaseModel(YamlModel):
    """
    Base Hoppr data model
    """

    class Config:  # pylint: disable=too-few-public-methods
        """Config for HopprBaseModel"""

        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        extra = Extra.forbid
        use_enum_values = True

    def __hash__(self) -> int:
        """
        Define to test equality or uniqueness between objects
        """
        return hash(repr(self))


class HopprBaseSchemaModel(HopprBaseModel):
    """
    Base Hoppr config file schema model
    """

    kind: Literal["Credentials", "Manifest", "Transfer"] = Field(..., description="Data model/schema kind")
    metadata: Metadata | None = Field(None, description="Metadata for the file")
    schema_version: str = Field(..., alias="schemaVersion", title="Schema Version")

    @validator("kind", allow_reuse=True, pre=True)
    @classmethod
    def validate_kind(cls, kind: str) -> str:
        """
        Return supplied `kind` value with only first letter capitalized
        """
        return kind.capitalize()



"""
Manifest file data model
"""
from __future__ import annotations

import uuid

from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, ClassVar, Literal, MutableMapping

from hoppr_cyclonedx_models.cyclonedx_1_4 import Component as ComponentSpecVersion14
from hoppr_cyclonedx_models.cyclonedx_1_4 import CyclonedxSoftwareBillOfMaterialsStandard as SbomSpecVersion14
from hoppr_cyclonedx_models.cyclonedx_1_4 import ExternalReference as ExternalReferenceSpecVersion14
from hoppr_cyclonedx_models.cyclonedx_1_4 import Property as PropertySpecVersion14
from packageurl import PackageURL
from pydantic import AnyUrl, Extra, Field, FileUrl, HttpUrl, NoneStr, create_model, validator
from pydantic.main import ModelMetaclass
from requests import HTTPError
from typer import secho

import hoppr.net
import hoppr.oci_artifacts
import hoppr.utils

from hoppr.constants import BomProps
from hoppr.exceptions import HopprLoadDataError
from hoppr.models.base import HopprBaseModel, HopprBaseSchemaModel
from hoppr.models.types import PurlType, RepositoryUrl

if TYPE_CHECKING:
    from pydantic.typing import DictStrAny


class Repository(HopprBaseModel):
    """
    Repository data model
    """

    url: RepositoryUrl
    description: NoneStr = None

    @validator("url", pre=True)
    @classmethod
    def validate_url(cls, url: str | RepositoryUrl) -> RepositoryUrl:
        """
        Validate URL string
        """
        if isinstance(url, str):
            url = RepositoryUrl(url=url)

        return url


# Dynamic Repositories model with PurlType values as attribute names
purl_type_repo_mapping = {str(purl_type): (list[Repository], Field([], unique_items=True)) for purl_type in PurlType}
RepositoriesMetaclass: ModelMetaclass = create_model(  # type: ignore[call-overload]
    "RepositoriesMetaclass", __base__=HopprBaseModel, **purl_type_repo_mapping
)


class Repositories(RepositoriesMetaclass):  # type: ignore[misc, valid-type]
    """
    Repositories data model
    """

    # pylint: disable=too-few-public-methods

    def __getitem__(self, item: str | PurlType) -> list[Repository]:
        return getattr(self, str(item), [])

    def __setitem__(self, item: str | PurlType, value: list[Repository]) -> None:
        setattr(self, str(item), value)

    def __iter__(self):
        for purl_type in PurlType:
            yield (purl_type, self[purl_type])


class LocalFile(HopprBaseModel):
    """
    LocalFile data model
    """

    local: Path


class OciFile(HopprBaseModel):
    """
    OciFile data model
    """

    oci: AnyUrl | str


class UrlFile(HopprBaseModel):
    """
    UrlFile data model
    """

    url: HttpUrl | FileUrl | AnyUrl | str


class SearchSequence(HopprBaseModel):
    """
    SearchSequence data model
    """

    version: Literal["v1"]
    repositories: list[HttpUrl | FileUrl | str] = []


IncludeRef = Annotated[LocalFile | UrlFile, Field(..., description="Reference to a local or remote manifest file")]
Includes = Annotated[list[IncludeRef], Field(..., description="List of manifest files to load")]
SbomRef = Annotated[LocalFile | OciFile | UrlFile, Field(..., description="Reference to a local or remote SBOM file")]
Sboms = Annotated[list[SbomRef], Field(..., description="List of SBOMs to process")]
SbomRefMap = Annotated[MutableMapping[SbomRef, "Sbom"], Field(...)]
ComponentPurlMap = Annotated[MutableMapping[str, "Component"], Field(...)]


class ExternalReference(HopprBaseModel, ExternalReferenceSpecVersion14):
    """
    ExternalReference data model derived from HopprBaseModel
    """

    class Config(HopprBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for ExternalReference model"
        extra = Extra.allow


class Property(HopprBaseModel, PropertySpecVersion14):
    """
    Property data model derived from HopprBaseModel
    """


class Component(HopprBaseModel, ComponentSpecVersion14):
    """
    Component data model derived from HopprBaseModel
    """

    class Config(HopprBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for Component model"
        extra = Extra.allow

    components: list[Component] | None = Field(None)  # type: ignore[assignment]
    externalReferences: list[ExternalReference] | None = Field(None)  # type: ignore[assignment]
    properties: list[Property] | None = Field(None)  # type: ignore[assignment]

    # Attributes not included in schema
    component_lookup: ClassVar[ComponentPurlMap] = {}

    @classmethod
    def find(cls, purl_string: str) -> Component | None:
        """
        Look up Component object by package URL string

        Args:
            purl_string (str): Package URL string to look up

        Returns:
            Component | None: Component object if found, otherwise None
        """
        return cls.component_lookup.get(purl_string)


class Sbom(HopprBaseModel, SbomSpecVersion14):
    """
    Sbom data model derived from HopprBaseModel
    """

    class Config(HopprBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for generic Sbom model"

        extra = Extra.allow

    components: list[Component] = Field(...)  # type: ignore[assignment]
    externalReferences: list[ExternalReference] | None = Field(None)  # type: ignore[assignment]

    # Attributes not included in schema
    loaded_sboms: ClassVar[SbomRefMap] = {}
    consolidated_sbom: ClassVar[Sbom]

    @classmethod
    def find(cls, ref_type: Literal["local", "oci", "url"], location: str | Path) -> Sbom | None:
        """
        Look up SBOM object by reference

        Args:
            ref_type (Literal["local", "oci", "url"]): Type of SBOM reference
            location (str | Path): Location of SBOM reference

        Returns:
            Sbom | None: SBOM object if found, otherwise None
        """
        match ref_type:
            case "local":
                return cls.loaded_sboms.get(LocalFile(local=Path(location)), None)
            case "oci":
                return cls.loaded_sboms.get(OciFile(oci=str(location)), None)
            case "url":
                return cls.loaded_sboms.get(UrlFile(url=str(location)), None)
            case _:
                return None


class ManifestFile(HopprBaseSchemaModel):
    """
    Data model to describe a single manifest file
    """

    kind: Literal["Manifest"]
    repositories: Repositories = Field(..., description="Maps supported PURL types to package repositories/registries")
    includes: Includes = []
    sboms: Sboms = []

    @classmethod
    def parse_file(cls, path: str | Path, *args, **kwargs) -> ManifestFile:  # pylint: disable=unused-argument
        """
        Override to resolve local file paths relative to manifest file
        """
        path = Path(path).resolve()

        data = hoppr.utils.load_file(path)
        if not isinstance(data, dict):
            raise TypeError("Local file content was not loaded as dictionary")

        # Resolve local file path references relative to manifest file path
        for sbom in data.get("sboms", []):
            if "local" in sbom:
                sbom["local"] = str((path.parent / sbom["local"]).resolve())

        for include in data.get("includes", []):
            if "local" in include:
                include["local"] = str((path.parent / include["local"]).resolve())

        return cls(**data)

    @classmethod
    def parse_obj(cls, obj: DictStrAny) -> ManifestFile:
        """
        Override to remove local file paths that can't be resolved
        """
        for include in list(obj.get("includes", [])):
            if "local" in include and not Path(include["local"]).is_absolute():
                secho(f"Skipping local include: relative path '{include['local']}' cannot be resolved", fg="yellow")
                obj["includes"].remove(include)

        for sbom in list(obj.get("sboms", [])):
            if "local" in sbom and not Path(sbom["local"]).is_absolute():
                secho(f"Skipping local SBOM: relative path '{sbom['local']}' cannot be resolved", fg="yellow")
                obj["sboms"].remove(sbom)

        return cls(**obj)


IncludeRefMap = Annotated[MutableMapping[IncludeRef, ManifestFile], Field(...)]


class Manifest(ManifestFile):
    """
    Manifest data model that generates lookups for `includes` and `sboms` references
    """

    # Attributes not included in schema
    loaded_manifests: ClassVar[IncludeRefMap] = {}

    @validator("includes", allow_reuse=True, always=True)
    @classmethod
    def populate_loaded_manifests(cls, includes: Includes, values: DictStrAny) -> Includes:
        """
        Validator that automatically loads manifest from local file or URL into lookup dictionary
        """
        for include_ref in includes:
            if include_ref not in cls.loaded_manifests:
                match include_ref:
                    case LocalFile():
                        cls.loaded_manifests[include_ref] = ManifestFile.parse_file(include_ref.local)
                        loaded = cls.load(include_ref.local)
                    case UrlFile():
                        data = hoppr.net.load_url(include_ref.url)
                        if not isinstance(data, dict):
                            raise TypeError("URL manifest include was not loaded as dictionary")

                        cls.loaded_manifests[include_ref] = ManifestFile.parse_obj(data)
                        loaded = cls.load(include_ref.url)
                    case _:
                        continue

                # Add all repositories of current manifest to included manifest
                for purl_type in PurlType:
                    # List of repos for purl type in current manifest
                    combined_repos = deepcopy(values["repositories"][str(purl_type)])

                    # List of repos for purl type in included manifest
                    include_repos: list[Repository] = getattr(loaded.repositories, str(purl_type), [])

                    # Append each included manifest repo to list of current manifest repos
                    for repo in include_repos:
                        if repo not in combined_repos:
                            combined_repos.append(repo)

                    # Set included manifest repos to the combined list of repos for purl type
                    setattr(loaded.repositories, str(purl_type), combined_repos)

                cls.loaded_manifests[include_ref] = loaded

        return includes

    @validator("sboms", allow_reuse=True, always=True)
    @classmethod
    def populate_loaded_sboms(cls, sboms: Sboms, values: DictStrAny) -> Sboms:
        """
        Validator that automatically loads SBOM from local file or URL into lookup dictionary
        """
        if not hasattr(Sbom, "consolidated_sbom"):
            Sbom.consolidated_sbom = Sbom(
                specVersion="1.4",
                version=1,
                bomFormat="CycloneDX",  # type: ignore[arg-type]
                serialNumber=uuid.uuid4().urn,
                components=[],
                externalReferences=[],
            )

        for sbom_ref in sboms:
            if sbom_ref not in Sbom.loaded_sboms:
                loaded = cls._load_sbom(sbom_ref)
                ref_url = cls._get_ref_url(sbom_ref)

                if loaded is None or ref_url is None:
                    continue

                loaded_sn = getattr(loaded, "serialNumber")

                # Merge current SBOM metadata into consolidated SBOM
                Sbom.consolidated_sbom.externalReferences.append(  # type: ignore[union-attr]
                    ExternalReference(
                        url=ref_url,
                        type="bom",  # type: ignore[arg-type]
                        comment=loaded_sn,
                        hashes=None,
                    )
                )

                for component in loaded.components or []:
                    purl_str = getattr(component, "purl", None)
                    if purl_str is None:
                        continue

                    purl_type = hoppr.utils.get_package_url(purl_str).type

                    if component.properties is None:
                        component.properties = []

                    # Generate the repository search sequence
                    search_sequence = SearchSequence(
                        version="v1", repositories=[str(repo.url) for repo in values["repositories"][purl_type]]
                    )

                    # Add repository search sequence as component property
                    component.properties.append(
                        Property(name=BomProps.COMPONENT_SEARCH_SEQUENCE, value=search_sequence.json())
                    )

                    # Add external reference to SBOM file that includes this component
                    if component.externalReferences is None:
                        component.externalReferences = []

                    component.externalReferences.append(
                        ExternalReference(
                            url=ref_url, comment=loaded_sn, type="bom", hashes=None  # type: ignore[arg-type]
                        )
                    )

                    # Merge component into consolidated SBOM
                    cls._add_component(component)

                Sbom.loaded_sboms[sbom_ref] = loaded

        return sboms

    @classmethod
    def _add_component(cls, component: Component) -> None:
        # Remove purl qualifiers for later comparison
        purl: PackageURL = hoppr.utils.get_package_url(component.purl)
        purl.qualifiers.clear()
        purl_str = purl.to_string()

        loaded = Component.find(purl_str)
        # Merge component into previously loaded component
        if loaded is not None:
            loaded.externalReferences = hoppr.utils.dedup_list(
                [*(loaded.externalReferences or []), *(component.externalReferences or [])]
            )

            if loaded.properties is None:
                loaded.properties = []  # pragma: no cover

            for prop in component.properties or []:
                if prop.name not in [loaded_prop.name for loaded_prop in loaded.properties]:
                    loaded.properties.append(prop)
        else:
            Sbom.consolidated_sbom.components.append(component)
            Component.component_lookup[purl_str] = component

    @classmethod
    def _get_ref_url(cls, sbom_ref: SbomRef) -> str | None:
        match sbom_ref:
            case LocalFile():
                return sbom_ref.local.as_uri()
            case OciFile():
                return sbom_ref.oci
            case UrlFile():
                return sbom_ref.url
            case _:
                return None

    @classmethod
    def _load_sbom(cls, sbom_ref: SbomRef) -> Sbom | None:
        match sbom_ref:
            case LocalFile():
                return Sbom.parse_file(sbom_ref.local)
            case OciFile():
                data = hoppr.oci_artifacts.pull_artifact(sbom_ref.oci)
                if not isinstance(data, dict):
                    raise TypeError("OCI URL SBOM file was not loaded as dictionary")

                return Sbom.parse_obj(data)
            case UrlFile():
                data = hoppr.net.load_url(sbom_ref.url)
                if not isinstance(data, dict):
                    raise TypeError("URL SBOM file was not loaded as dictionary")

                return Sbom.parse_obj(data)
            case _:
                return None

    @classmethod
    def find(cls, ref_type: Literal["local", "url"], location: str | Path) -> ManifestFile | None:
        """
        Lookup manifest object by include reference

        Args:
            ref_type (Literal["local", "url"]): Type of include
            location (str | Path): Path to included manifest

        Returns:
            ManifestFile | None: Manifest object if found, otherwise None
        """
        match ref_type:
            case "local":
                return cls.loaded_manifests.get(LocalFile(local=Path(location)))
            case "url":
                return cls.loaded_manifests.get(UrlFile(url=str(location)))
            case _:
                return None

    @classmethod
    def load(cls, source: str | Path | DictStrAny) -> Manifest:
        """
        Load manifest from local file, URL, or dict
        """
        match source:
            case dict():
                data = source
            case Path():
                path = Path(source).resolve()
                manifest_file = cls.parse_file(path)
                local_ref = LocalFile(local=path)
                cls.loaded_manifests[local_ref] = manifest_file
                data = manifest_file.dict(by_alias=True)
            case str():
                try:
                    include_dict = hoppr.net.load_url(source)
                    if not isinstance(include_dict, dict):
                        raise TypeError("URL manifest include was not loaded as dictionary")

                    manifest_file = cls.parse_obj(include_dict)
                    url_ref = UrlFile(url=source)
                    cls.loaded_manifests[url_ref] = manifest_file
                except (HopprLoadDataError, HTTPError) as ex:
                    raise HopprLoadDataError from ex

                data = manifest_file.dict(by_alias=True)

        return cls(**data)



"""
Collector plugin to copy artifacts using the Nexus API
"""
from __future__ import annotations

import json
import re

from pathlib import Path
from time import sleep
from typing import Any
from urllib.parse import urlparse

import requests

from packageurl import PackageURL  # type: ignore
from requests.auth import HTTPBasicAuth

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models import HopprContext
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.types import PurlType, RepositoryUrl
from hoppr.net import download_file
from hoppr.result import Result


class CollectNexusSearch(SerialCollectorPlugin):
    """
    Class to copy artifacts using the Nexus API
    """

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)
        if self.config is not None:
            if "purl_types" in self.config:
                self.supported_purl_types = self.config["purl_types"]

    def supports_purl_type(self, purl_type: str) -> bool:
        """
        Indicates whether or not this particular plug-in supports components of the specified
        purl type.

        Unless specified by the user, the Nexus search collector supports all types EXCEPT those
        listed below
        """

        if self.supported_purl_types:
            return super().supports_purl_type(purl_type)

        return purl_type not in ["git", "github", "gitlab"]

    @hoppr_rerunner
    def collect(self, comp: Any, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        """
        Collect artifact from Nexus
        """

        auth: HTTPBasicAuth | None = None
        if creds is not None:
            auth = HTTPBasicAuth(username=creds.username, password=creds.password.get_secret_value())

        nexus_url = self._parse_nexus_url(repo_url)[0]

        if not CollectNexusSearch.is_nexus_instance(nexus_url, auth):
            return Result.fail(f"{nexus_url} is not a Nexus instance")

        purl: PackageURL = PackageURL.from_string(comp.purl)  # pyright: ignore[reportGeneralTypeIssues]

        source_urls = CollectNexusSearch.get_download_urls(purl, repo_url, auth)
        if len(source_urls) == 0:
            msg = f"No artifacts found in Nexus instance {repo_url} for purl {comp.purl}"
            self.get_logger().error(msg, indent_level=2)
            return Result.fail(msg)

        target_dir = self._directory_for_nexus(purl, source_urls[0])

        for source_url in source_urls:
            self.get_logger().info(
                msg=f"Collecting from {source_url}",
                indent_level=2,
            )

            file_name = Path(urlparse(source_url).path).parts[-1]

            response = download_file(source_url, str(target_dir / file_name))
            nexus_result = Result.from_http_response(response)

            self.get_logger().info(f"Download Result: {nexus_result}", indent_level=3)

            if not nexus_result.is_success():
                return nexus_result

        self.set_collection_params(comp, repo_url, target_dir)
        return Result.success(return_obj=comp)

    @staticmethod
    def _parse_nexus_url(repo_url: str) -> tuple[str, str | None]:
        nexus_url = repo_url
        nexus_repo = None
        repo_specified = re.search(r"(https?://.*?)/repository/(.*?)(/.*)?$", repo_url)
        if repo_specified:
            nexus_url = repo_specified.group(1)
            nexus_repo = repo_specified.group(2)

        return (nexus_url, nexus_repo)

    def _directory_for_nexus(self, purl: PackageURL, url: str) -> Path:
        repo_match = re.search(r"(.*?/repository/.*?)(/.*)?/(.*)", url)
        if repo_match:
            nexus_repo = repo_match.group(1)
            path = repo_match.group(2)
            if path is not None:
                path = path[1:]

        subdir = None
        match purl.type:
            case "docker" | "generic" | "maven" | "rpm":
                subdir = path
            case "helm" | "pypi":
                subdir = f"{purl.name}_{purl.version}"

        target_dir = self.directory_for(purl.type, nexus_repo, subdir=subdir)

        return target_dir

    @staticmethod
    def is_nexus_instance(repo_url: str, auth: HTTPBasicAuth | None = None) -> bool:
        """
        Checks whether or not the repo_url refers to a Nexus instance
        """
        test_url = RepositoryUrl(url=repo_url) / "service" / "rest" / "v1" / "status"

        for attempt in range(3):
            if attempt > 0:
                sleep(5)

            response = requests.get(
                f"{test_url}",
                auth=auth,
                allow_redirects=True,
                stream=True,
                verify=True,
                timeout=60,
            )

            if response.status_code < 300:
                return True
            if response.status_code < 500:
                return False

        return False

    @staticmethod
    def get_download_urls(purl: PackageURL, repo_url: str, auth: HTTPBasicAuth | None = None) -> list[str]:
        """
        Retrieves all urls to be retrieved from Nexus for this component
        """
        nexus_url, nexus_repo = CollectNexusSearch._parse_nexus_url(repo_url)
        search_url = RepositoryUrl(url=nexus_url) / "service" / "rest" / "v1" / "search" / "assets"

        additional_search_params: list[dict] = [{}]
        nexus_format = purl.type
        match purl.type:
            case "deb":
                nexus_format = "apt"
            case "gems":
                nexus_format = "rubygems"
            case "golang":
                nexus_format = "go"
            case "generic":
                nexus_format = "raw"
            case "maven":
                nexus_format = "maven2"
                additional_search_params = [
                    {"maven.extension": "jar", "maven.classifier": ""},
                    {"maven.extension": "jar", "maven.classifier": "sources"},
                    {"maven.extension": "pom"},
                ]
            case "rpm" | "yum":
                nexus_format = "yum"
                if purl.qualifiers.get("arch") is not None:
                    additional_search_params = [{"yum.architecture": purl.qualifiers.get('arch')}]

        base_params = {"sort": "version", "name": purl.name, "format": nexus_format}

        if purl.version is not None:
            base_params["version"] = purl.version

        if nexus_repo is not None:
            base_params["repository"] = nexus_repo

        url_list = []
        for extra_search_parms in additional_search_params:
            response = requests.get(
                f"{search_url}",
                auth=auth,
                allow_redirects=True,
                stream=True,
                verify=True,
                timeout=60,
                params=base_params | extra_search_parms,
            )

            if response.status_code < 300:
                search_result = json.loads(response.content)
                for item in search_result["items"]:
                    if item["downloadUrl"].startswith(repo_url):
                        url_list.append(item["downloadUrl"])
                        break

        return url_list

    @classmethod
    def get_attestation_products(cls, config: dict | None = None) -> list[str]:
        products = []
        if config is not None and "purl_types" in config:
            for purl_type in config["purl_types"]:
                products.append(purl_type + "/*")

        else:
            for purl_type in PurlType:
                if str(purl_type) not in ["git", "github", "gitlab"]:
                    products.append(str(purl_type) + "/*")

        return products



"""Functions to faciliate interaction with non-image OCI Artifacts"""
import re

from pathlib import Path
from typing import Dict, Tuple

from oras.client import OrasClient  # type: ignore[import]
from oras.container import Container  # type: ignore[import]
from oras.provider import Registry  # type: ignore[import]

from hoppr import utils
from hoppr.exceptions import HopprLoadDataError
from hoppr.models.credentials import Credentials

# Regex per: https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
SEM_VER_CORE_REGEX = r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)$"
SEM_VER_FULL_REGEX = (
    r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)"
    r"(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
)


def pull_artifact(artifact: str, allow_version_discovery=False, version_regex=SEM_VER_CORE_REGEX) -> Dict:
    """
    Pulls OCI Artifact from registry into memory and returns content

    If no version (tag) is provided on URL, hoppr can attempt to disover highest version if allow_version_discovery
    is enabled. A regex for match versions can be provided to handle less standard version formats.
    Default will look for SemVer2 core
    """
    pulled = _pull_artifact(artifact, allow_version_discovery=allow_version_discovery, version_regex=version_regex)

    # We should only get one. If we get more, just take the first
    loaded_content = utils.load_file(pulled)
    if not isinstance(loaded_content, dict):
        raise HopprLoadDataError("Pulled artifact data was not loaded as dictionary")

    return loaded_content


def pull_artifact_to_disk(
    artifact: str,
    target_dir: str,
    allow_version_discovery=False,
    version_regex=SEM_VER_CORE_REGEX,
):
    """Pulls OCI Artifact from registry to specified target_dir.

    If no version (tag) is provided on URL, hoppr can attempt to disover highest version if allow_version_discovery
    is enabled. A regex for match versions can be provided to handle less standard version formats.
    Default will look for SemVer2 core
    """
    return _pull_artifact(
        artifact,
        target_dir,
        allow_version_discovery=allow_version_discovery,
        version_regex=version_regex,
    )


def _create_client_info(artifact: str) -> Tuple[Container, OrasClient]:
    """Utility function for creating ORAS client and intermediate objects"""
    # Use ORAS' Container object to faciliate parsing registry host
    container = Container(artifact)
    registry = Registry(hostname=container.registry)

    creds = Credentials.find(container.registry)
    if creds is not None:
        registry.set_basic_auth(creds.username, creds.password.get_secret_value())
    client = OrasClient(registry=registry)

    return container, client


def _pull_artifact(
    artifact: str,
    outdir=None,
    allow_version_discovery=False,
    version_regex=SEM_VER_CORE_REGEX,
) -> Path:
    """Looks up and pulls artifact to disk from OCI registry"""
    container, client = _create_client_info(artifact)

    full_artifact = artifact
    # Only discover tag if it hasn't been explicitly defined.
    # We let the 'Container' do the parsing, so we'll just compare
    if not artifact.endswith(container.tag):
        if not allow_version_discovery:
            raise HopprLoadDataError(
                f"No version provided and version discovery is disabled. Cannot pull artifact {artifact}"
            )
        version_tag = _discover_highest_version_tag(artifact, version_regex=version_regex)
        full_artifact += f":{version_tag}"

    files = client.pull(target=full_artifact, outdir=outdir)
    return Path(files[0])


def _discover_highest_version_tag(artifact: str, version_regex=SEM_VER_CORE_REGEX):
    """Attempts to discover the highest versioned artifact based on provided
    regex. Default will use full SemVer"""
    container, client = _create_client_info(artifact)

    tags: list[str] = client.get_tags(str(container))

    # Only find those that match semver
    # Future enhancement can allow user to define regex to use in the artifact definition
    tags = sorted(filter(lambda x: re.match(version_regex, x), tags["tags"]))  # type: ignore[arg-type,call-overload]

    return sorted(tags)[-1]



"""
Plugin to remove SBOM components that are specified by a "previous" SBOM
"""
import io
import tarfile

from copy import deepcopy
from pathlib import Path

from packageurl import PackageURL

from hoppr import __version__
from hoppr.base_plugins.hoppr import HopprPlugin, hoppr_process
from hoppr.exceptions import HopprError
from hoppr.models.manifest import Component, Manifest, Sbom
from hoppr.models.types import BomAccess
from hoppr.result import Result
from hoppr.utils import dedup_list, get_package_url, load_string


class DeltaSbom(HopprPlugin):
    """
    Plugin to remove SBOM components that are specified by a "previous" SBOM
    """

    bom_access = BomAccess.FULL_ACCESS
    products: list[str] = ["generic/_metadata_/_previous_bom.json"]

    def get_version(self) -> str:
        return __version__

    @hoppr_process
    def pre_stage_process(self):
        """
        Tar-up the context.collect_root_dir directory
        """

        if self.context.previous_delivery is not None:
            previous_source = str(self.context.previous_delivery)
        elif self.config is None or self.config.get("previous") is None:
            return Result.success(
                "No previously delivered bundle specified for delta bundle. All components will be delivered"
            )
        else:
            previous_source = self.config.get("previous")

        if not Path(previous_source).exists():  # pyright: ignore[reportGeneralTypeIssues]
            return Result.fail(f"Previous source file \"{previous_source}\" not found.")

        self.get_logger().info(f"Creating delta/update SBOM, previous SBOM being retrieved from {previous_source}")

        previous_sbom = self._get_previous_bom(previous_source)  # pyright: ignore[reportGeneralTypeIssues]

        target_dir = self.context.collect_root_dir / "generic" / "_metadata_"
        target_dir.mkdir(parents=True, exist_ok=True)

        with (target_dir / "_previous_bom.json").open(mode="w", encoding="utf-8") as bom_data:
            bom_data.write(previous_sbom.json(exclude_none=True, by_alias=True, indent=2))

        delta_sbom = deepcopy(self.context.delivered_sbom)
        delta_sbom.components = []

        for new_comp in self.context.delivered_sbom.components or []:
            include_component = True
            for prev_comp in previous_sbom.components or []:
                if DeltaSbom._component_match(new_comp, prev_comp):
                    include_component = False
                    break

            if include_component:
                self.get_logger().debug(f"Including purl {new_comp.purl}", indent_level=1)
                delta_sbom.components.append(new_comp)

        self.get_logger().info(f"Input sbom has {len(self.context.delivered_sbom.components)} components")
        self.get_logger().info(f"Prev  sbom has {len(previous_sbom.components)} components")
        self.get_logger().info(f"Delta sbom has {len(delta_sbom.components)} components")

        if len(delta_sbom.components) == 0:
            return Result.fail(f"No components updated since \"{previous_source}\".")

        return Result.success(
            f"Delivering updates for {len(delta_sbom.components)} of "
            f"{len(self.context.delivered_sbom.components)} components.",
            return_obj=delta_sbom,
        )

    @staticmethod
    def _get_previous_bom(source: str) -> Sbom:
        try:
            Manifest.load(Path(source))
            return Sbom.consolidated_sbom
        except (TypeError, UnicodeDecodeError):
            pass

        with tarfile.open(source) as tar:
            buffer = tar.extractfile("./generic/_metadata_/_consolidated_bom.json")

            if buffer is None:
                raise HopprError("Unable to extract BOM file from tar")

            with io.TextIOWrapper(buffer) as bom_file:
                content: str = bom_file.read()
                bom_dict = load_string(content)
                if not isinstance(bom_dict, dict):
                    raise HopprError("Invalid BOM file retrieved from tar")
                return Sbom(**bom_dict)

    @staticmethod
    def _purl_match(new_purl: PackageURL, prev_purl: PackageURL) -> bool:

        if (
            new_purl.name != prev_purl.name
            or new_purl.type != prev_purl.type
            or (new_purl.namespace or "") != (prev_purl.namespace or "")
            or (new_purl.version or "") != (prev_purl.version or "")
            or (new_purl.subpath or "") != (prev_purl.subpath or "")
        ):
            return False

        qual_keys = list(new_purl.qualifiers.keys()) or []
        qual_keys.extend(list(prev_purl.qualifiers.keys()) or [])

        for key in dedup_list(qual_keys):
            if new_purl.qualifiers.get(key) != prev_purl.qualifiers.get(key):
                return False

        return True

    @staticmethod
    def _component_match(new_comp: Component, prev_comp: Component) -> bool:

        new_purl: PackageURL = get_package_url(new_comp.purl)
        prev_purl: PackageURL = get_package_url(prev_comp.purl)

        if not DeltaSbom._purl_match(new_purl, prev_purl):
            return False

        hash_matches = 0

        for new_hash in new_comp.hashes or []:
            for prev_hash in prev_comp.hashes or []:
                if new_hash.alg.value != prev_hash.alg.value:
                    continue
                if new_hash.content != prev_hash.content:
                    return False
                hash_matches += 1

        if hash_matches > 0:
            return True

        if new_purl.version is None or new_purl.version == "latest":
            return False

        return True



"""
Hoppr utility functions
"""
from __future__ import annotations

import importlib
import inspect

from abc import ABCMeta
from functools import cache
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, Type

from packageurl import PackageURL
from ruamel.yaml import YAML
from ruamel.yaml.parser import ParserError as YAMLParserError
from ruamel.yaml.scanner import ScannerError as YAMLScannerError

from hoppr.exceptions import HopprLoadDataError, HopprPluginError

if TYPE_CHECKING:
    from hoppr.base_plugins.hoppr import HopprPlugin
    from hoppr.models import HopprContext


def _class_in_module(cls_type: ABCMeta, plugin_module: ModuleType):
    """
    Determines if the specified object is a class
    defined in the module
    """
    return inspect.isclass(cls_type) and cls_type.__module__ == plugin_module.__name__


def plugin_class(plugin_name: str) -> Type[HopprPlugin]:
    """
    return a concrete class of an object defined by a plugin name

    Assumes the specified plugin will define exactly one concrete class, which
    will be instaniated using a default constructor (i.e., one with no parameters).
    """
    try:
        plugin_module = importlib.import_module(name=plugin_name)
    except ModuleNotFoundError as mnfe:
        raise ModuleNotFoundError(f"Unable to locate plug-in {plugin_name}: {mnfe}") from mnfe

    plugin_cls = None

    for cls_name, cls_type in inspect.getmembers(plugin_module, inspect.isclass):
        if _class_in_module(cls_type, plugin_module):
            if plugin_cls is not None:
                raise HopprPluginError(
                    f"Multiple candidate classes defined in {plugin_name}: {plugin_cls.__name__}, {cls_name}"
                )

            plugin_cls = cls_type

    if plugin_cls is None:
        raise HopprPluginError(f"No class definition found in in {plugin_name}.")

    return plugin_cls


def plugin_instance(plugin_name: str, context: HopprContext, config: Any = None):
    """
    Create an instance of an object defined by a plugin name

    Assumes the specified plugin will define exactly one concrete class, which
    will be instaniated using a default constructor (i.e., one with no parameters).
    """
    plugin_cls = plugin_class(plugin_name)

    return plugin_cls(context=context, config=config)


def load_string(contents: str) -> dict | list | None:
    """
    Return a YAML or JSON formatted string as a dict
    """
    if not contents.strip():
        raise HopprLoadDataError("Empty string cannot be parsed.")

    # Replace tab characters with spaces to prevent parsing errors
    contents = contents.replace("\t", "  ")
    loaded_contents: dict | list | None = None

    try:
        # Applicable to both YAML and JSON formats since valid JSON data is also valid YAML
        yaml = YAML(typ="safe", pure=True)
        contents = contents.replace("\t", "  ")
        loaded_contents = yaml.load(contents)

        # yaml.safe_load will sometimes return a single string rather than the required structure
        if isinstance(loaded_contents, str):
            raise HopprLoadDataError("Expected dictionary or list, but contents were loaded and returned as string")
    except (YAMLParserError, YAMLScannerError) as ex:
        raise HopprLoadDataError("Unable to recognize data as either json or yaml") from ex
    except HopprLoadDataError as ex:
        raise HopprLoadDataError from ex

    return loaded_contents


def load_file(input_file_path: Path) -> dict | list | None:
    """
    Load file content (either JSON or YAML) into a dict
    """

    if not input_file_path.is_file():
        raise HopprLoadDataError(f"{input_file_path} is not a file, cannot be opened.")

    with input_file_path.open(mode="r", encoding="utf-8") as input_file:
        content: str = input_file.read()
        if not content.strip():
            raise HopprLoadDataError(f"File {input_file_path} is empty.")

    return load_string(content)


def dedup_list(list_in: list[Any]) -> list[Any]:
    """
    De-duplicate a list
    """
    return list(dict.fromkeys(list_in)) if list_in is not None else []


def obscure_passwords(command_list: list[str], sensitive_args: list[str] | None = None) -> str:
    """
    Returns an input string with any specified passwords hidden
    """

    password_list: list[str] = sensitive_args if sensitive_args is not None else []
    obscured_command_list: list[str] = []

    for arg in command_list:
        # Quote arguments that contain spaces
        if " " in arg:
            arg = f'"{arg}"'

        # Replace password string(s) in argument
        for password in password_list:
            if password is not None and len(password) > 0:
                arg = arg.replace(password, "[masked]")

        obscured_command_list.append(arg)

    return " ".join(obscured_command_list)


def remove_empty(directory: Path) -> set[Path]:
    """
    Removes empty folders given the directory including parent folders
    """
    deleted: set[Path] = set()

    if not directory.exists():
        raise FileNotFoundError()

    for subdir in directory.iterdir():
        if subdir.is_dir():
            deleted.update(remove_empty(subdir))

    if directory.is_dir() and not any(directory.iterdir()):
        directory.rmdir()
        deleted.add(directory)

    return deleted


@cache
def get_package_url(purl_string: str) -> PackageURL:
    """
    Get the PackageURL for a given purl_string and store it in a cache for improved performance.

    Args:
        purl_string (str): The string representation of a Package URL

    Returns:
        PackageURL: The object representation of a PackageURL

    """
    return PackageURL.from_string(purl_string)  # pyright: ignore[reportGeneralTypeIssues]



"""
Collector plugin for maven artifacts
"""
from __future__ import annotations

import os

from pathlib import Path
from subprocess import CalledProcessError
from tempfile import NamedTemporaryFile
from typing import Any, OrderedDict

import jmespath
import xmltodict

from packageurl import PackageURL

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models import HopprContext
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.manifest import Component
from hoppr.result import Result

_MAVEN_DEP_PLUGIN = "org.apache.maven.plugins:maven-dependency-plugin:3.5.0"


class CollectMavenPlugin(SerialCollectorPlugin):
    """
    Collector plugin for maven artifacts
    """

    supported_purl_types = ["maven"]
    required_commands = ["mvn"]
    products: list[str] = ["maven/*"]
    system_repositories: list[str] = ["https://repo.maven.apache.org/maven2"]

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)

        self.extra_opts: list[str] = []

        if self.config is not None:
            self.required_commands = self.config.get("maven_command", self.required_commands)
            self.extra_opts = self.config.get("maven_opts", self.extra_opts)

        system_settings_file = Path("/") / "etc" / "maven" / "settings.xml"
        user_settings_file = Path.home() / ".m2" / "settings.xml"

        if not self.context.strict_repos:

            # Identify system repositories
            for settings_file in [system_settings_file, user_settings_file]:
                if settings_file.is_file():
                    settings_dict: OrderedDict[str, Any] = xmltodict.parse(
                        settings_file.read_text(encoding="utf-8"),
                        encoding="utf-8",
                        force_list={"profile", "repository"},
                    )

                    repo_urls: list[str] = jmespath.search(
                        expression="settings.profiles.profile[].repositories.repository[].url", data=settings_dict
                    )

                    for repo in repo_urls:
                        if repo not in self.system_repositories:
                            self.system_repositories.append(repo)

    def _get_maven_component(self, command: list[str], password_list: list[str], **kwargs):
        full_command = command.copy()
        for key, value in kwargs.items():
            full_command.append(f"-D{key}={value}")

        result = self.run_command(full_command, password_list)

        # The maven plugin does not recognize the 'destFileName' argument, so rename the file
        if result.returncode == 0:
            _, name, version, extension = kwargs.get("artifact", "::::").split(":")
            directory = kwargs.get("outputDirectory", ".")
            old_name = f"{name}-{version}.{extension}"
            new_name = kwargs.get("destFileName", f"{name}_{version}.{extension}")

            os.rename(directory / old_name, directory / new_name)

        return result

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        """
        Copy a component to the local collection directory structure
        """
        # pylint: disable=too-many-locals

        purl: PackageURL = PackageURL.from_string(comp.purl)  # pyright: ignore[reportGeneralTypeIssues]
        artifact = f"{purl.namespace}:{purl.name}:{purl.version}"

        extension = purl.qualifiers.get("type", "tar.gz")
        target_dir = self.directory_for(purl.type, repo_url, subdir=f"{purl.namespace}")

        self.get_logger().info(msg="Copying maven artifact:", indent_level=2)
        self.get_logger().info(msg=f"source: {repo_url}", indent_level=3)
        self.get_logger().info(msg=f"destination: {target_dir}", indent_level=3)

        settings_dict = {
            "settings": {
                "servers": {
                    "server": {
                        "id": "repoId",
                        "username": "${repo.login}",
                        "password": "${repo.pwd}",
                    }
                }
            }
        }

        with NamedTemporaryFile(mode="w+", encoding="utf-8") as settings_file:
            settings_file.write(xmltodict.unparse(input_dict=settings_dict, pretty=True))
            settings_file.flush()

            password_list = []

            defines = {
                "artifact": f"{artifact}:{extension}",
                "outputDirectory": target_dir,
                "destFileName": f"{purl.name}_{purl.version}.{extension}",
                "remoteRepositories": f"repoId::::{repo_url}",
            }

            if creds is not None:
                defines["repo.login"] = creds.username
                defines["repo.pwd"] = creds.password.get_secret_value()
                password_list = [creds.password.get_secret_value()]

            command = [
                self.required_commands[0],
                f"{_MAVEN_DEP_PLUGIN}:copy",
                f"--settings={settings_file.name}",
                *self.extra_opts,
            ]

            run_result = self._get_maven_component(command, password_list, **defines)
            error_msg = f"Failed to download maven artifact {artifact} type={extension}"

            try:
                run_result.check_returncode()

                if extension != "pom":
                    defines["artifact"] = str(defines["artifact"]).replace(f":{extension}", ":pom")
                    defines["destFileName"] = str(defines["destFileName"]).replace(f".{extension}", ".pom")

                    error_msg = f"Failed to download pom for maven artifact {artifact}"
                    run_result = self._get_maven_component(command, password_list, **defines)
                    run_result.check_returncode()
            except CalledProcessError:
                self.get_logger().debug(msg=error_msg, indent_level=2)
                return Result.retry(message=error_msg)

        self.set_collection_params(comp, repo_url, target_dir)

        return Result.success(return_obj=comp)



"""
Collector plugin for docker images
"""
from __future__ import annotations

import os
import re
import urllib.parse

from typing import Any

from packageurl import PackageURL

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models import HopprContext
from hoppr.models.credentials import CredentialRequiredService
from hoppr.result import Result


class CollectDockerPlugin(SerialCollectorPlugin):
    """
    Collector plugin for docker images
    """

    supported_purl_types = ["docker"]
    required_commands = ["skopeo"]
    products: list[str] = ["docker/*"]
    system_repositories: list[str] = ["https://docker.io/"]
    process_timeout = 300

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)
        if self.config is not None:
            if "skopeo_command" in self.config:
                self.required_commands = [self.config["skopeo_command"]]

    @hoppr_rerunner
    def collect(self, comp: Any, repo_url: str, creds: CredentialRequiredService | None = None):
        """
        Copy a component to the local collection directory structure
        """
        purl = PackageURL.from_string(comp.purl)

        source_image = os.path.join(repo_url, purl.namespace or "", purl.name + ":" + purl.version)
        source_image = re.sub(r"^https?://", "", source_image)

        file_name = urllib.parse.unquote(purl.name) + "_" + purl.version
        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)
        target_path = os.path.join(target_dir, file_name)
        destination = f'docker-archive:{target_path}:{source_image}'

        if not source_image.startswith("docker://"):
            source_image = "docker://" + source_image

        self.get_logger().info(msg="Copying docker image:", indent_level=2)
        self.get_logger().info(msg=f"source: {source_image}", indent_level=3)
        self.get_logger().info(msg=f"destination: {destination}", indent_level=3)

        command = [self.required_commands[0], "copy"]

        password_list = []

        if creds is not None:
            password_list = [creds.password.get_secret_value()]
            command.extend(["--src-creds", f"{creds.username}:{creds.password.get_secret_value()}"])

        if re.match("^http://", repo_url):
            command.append("--src-tls-verify=false")

        command.extend([source_image, destination])

        proc = self.run_command(command, password_list)

        if proc.returncode != 0:
            msg = f"Skopeo failed to copy docker image to {destination}, " + f"return_code={proc.returncode}"

            self.get_logger().debug(msg=msg, indent_level=2)

            if os.path.exists(target_path):
                self.get_logger().info(
                    msg="Artifact collection failed, deleting file and retrying",
                    indent_level=2,
                )
                os.remove(target_path)

            return Result.retry(message=msg)

        self.set_collection_params(comp, repo_url, target_dir)

        return Result.success(return_obj=comp)



"""
Hoppr Network utility functions
"""
import requests

from requests.auth import HTTPBasicAuth

from hoppr.exceptions import HopprLoadDataError
from hoppr.models.credentials import CredentialRequiredService, Credentials
from hoppr.utils import load_string


def load_url(url: str):
    """
    Load config content (either json or yml) from a url into a dict
    """
    creds = Credentials.find(url)

    response = None
    if creds is not None:
        authorization_headers = {
            "PRIVATE-TOKEN": creds.password.get_secret_value(),
            "Authorization": f"Bearer {creds.password.get_secret_value()}",
        }

        basic_auth = HTTPBasicAuth(username=creds.username, password=creds.password.get_secret_value())
        response = requests.get(url, auth=basic_auth, headers=authorization_headers, timeout=60)
    else:
        response = requests.get(url, timeout=60)

    response.raise_for_status()
    valid_data = True
    try:
        if isinstance(response.content, bytes):
            return load_string(response.content.decode("utf-8"))
        if isinstance(response.content, str):
            return load_string(response.content)
        valid_data = False
    except HopprLoadDataError as parse_error:
        message = f"Unable to parse result from {url}."
        if response.url != url:
            message += f" Request was redirected to {response.url}. An auth issue might have occurred."
        raise HopprLoadDataError(message) from parse_error

    if not valid_data:
        raise HopprLoadDataError("Response type is not bytes or str")

    return None  # pragma: no cover


def download_file(url: str, dest: str, creds: CredentialRequiredService | None = None):
    """
    Download content from a url into a file
    """
    if creds is None:
        creds = Credentials.find(url)

    basic_auth = None
    if creds is not None:
        basic_auth = HTTPBasicAuth(username=creds.username, password=creds.password.get_secret_value())

    response = requests.get(url, auth=basic_auth, allow_redirects=True, stream=True, verify=True, timeout=60)

    if 200 <= response.status_code < 300:
        with open(dest, "wb") as out_file:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    out_file.write(chunk)

    return response



"""
Collector plugin for raw files
"""

import shutil

from pathlib import Path
from typing import Any, List
from urllib.parse import unquote

from packageurl import PackageURL

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.types import RepositoryUrl
from hoppr.net import download_file
from hoppr.result import Result


class CollectRawPlugin(SerialCollectorPlugin):
    """
    Collector plugin for raw files
    """

    supported_purl_types = ["binary", "generic", "raw"]
    products: List[str] = ["binary/*", "generic/*", "raw/*"]

    def get_version(self) -> str:
        return __version__

    @hoppr_rerunner
    def collect(self, comp: Any, repo_url: str, creds: CredentialRequiredService | None = None):
        """
        Copy a component to the local collection directory structure
        """
        source_url = RepositoryUrl(url=repo_url)

        purl = PackageURL.from_string(comp.purl)

        subdir = None
        if purl.namespace is not None:
            source_url /= f"{purl.namespace}"
            subdir = unquote(purl.namespace)

        target_dir = self.directory_for(purl.type, repo_url, subdir=subdir)

        file_name = unquote(purl.name)

        if source_url.scheme == "file":
            repo_path = Path(repo_url.removeprefix("file:").removeprefix("//"))
            source_file = (repo_path / (purl.namespace or "") / file_name).expanduser()

            self.get_logger().info(msg="Copying component:", indent_level=2)
            self.get_logger().info(msg=f"source: {source_file}", indent_level=3)
            self.get_logger().info(msg=f"destination: {target_dir.joinpath(file_name)}", indent_level=3)

            if not source_file.is_file():
                msg = f"Unable to locate file {source_file}, skipping remaining attempts"
                self.get_logger().error(msg=msg, indent_level=2)
                return Result.fail(message=msg)

            shutil.copy(source_file, target_dir)

            self.set_collection_params(comp, repo_url, target_dir)
            return Result.success(return_obj=comp)

        download_url = source_url / file_name

        self.get_logger().info(msg="Downloading file:", indent_level=2)
        self.get_logger().info(msg=f"source: {download_url}", indent_level=3)
        self.get_logger().info(msg=f"destination: {target_dir.joinpath(file_name)}", indent_level=3)

        response = download_file(f"{download_url}", str(target_dir / file_name), creds)
        result = Result.from_http_response(response)

        if result.is_fail():
            msg = f"Unable to download from {download_url}, skipping remaining attempts"
            self.get_logger().error(msg=msg, indent_level=2)

        if not result.is_success():
            return result

        self.set_collection_params(comp, repo_url, target_dir)

        return Result.success(return_obj=comp)



"""
Run Hoppr processing, using multiple processors
"""
from __future__ import annotations

import logging
import multiprocessing
import os
import shutil
import socket
import sys
import tempfile
import time
import uuid

from concurrent.futures import Future, ProcessPoolExecutor
from copy import deepcopy
from datetime import datetime
from multiprocessing import cpu_count
from os import PathLike
from pathlib import Path
from threading import _RLock as RLock
from typing import Any, MutableMapping
from urllib.parse import quote_plus

import jmespath

from packageurl import PackageURL
from typer import colors, echo, secho

from hoppr import __version__
from hoppr.base_plugins.hoppr import HopprPlugin
from hoppr.core_plugins.report_generator import Report, ReportGenerator
from hoppr.exceptions import HopprPluginError
from hoppr.in_toto import HopprInTotoLinks
from hoppr.mem_logger import MemoryLogger
from hoppr.models import HopprContext
from hoppr.models.credentials import Credentials
from hoppr.models.manifest import Component, Manifest, ManifestFile, Sbom
from hoppr.models.transfer import ComponentCoverage, Plugin, StageRef, Transfer
from hoppr.models.types import BomAccess
from hoppr.result import Result
from hoppr.utils import plugin_class, plugin_instance

# Import on Unix systems only
if os.name == "posix":
    import pwd


def _run_plugin(
    plugin_name: str,
    context: HopprContext,
    config: dict[str, Any] | None,
    method_name: str,
    component: Component | None,
) -> Result:
    """
    Runs a single method for a single component (if supplied) on a single plugin
    """
    plugin = plugin_instance(plugin_name, context, config)

    match method_name:
        case HopprPlugin.pre_stage_process.__name__:
            result = plugin.pre_stage_process()
        case HopprPlugin.process_component.__name__:
            result = plugin.process_component(component)
        case HopprPlugin.post_stage_process.__name__:
            result = plugin.post_stage_process()
        case _:
            result = Result.fail(f"Invalid method: {method_name}")

    if result.return_obj is not None:
        if not plugin.bom_access.has_access_to(result.return_obj):
            result = Result.fail(
                f"Plugin {type(plugin).__name__} has BOM access level {plugin.bom_access.name}, "
                f"but returned an object of type {type(result.return_obj).__name__}"
            )

    return result


class StageProcessor:  # pylint: disable=too-few-public-methods
    """
    Class to handle all processing within a single Hoppr stage
    """

    component_based_methods = [HopprPlugin.process_component.__name__]

    def __init__(self, stage_ref: StageRef, context: HopprContext):
        self.stage_id = stage_ref.name
        self.context = context
        self.plugin_ref_list = stage_ref.plugins
        self.config_component_coverage = None
        if stage_ref.component_coverage is not None:
            self.config_component_coverage = ComponentCoverage[str(stage_ref.component_coverage)]
        self.required_coverage = ComponentCoverage.OPTIONAL
        self.results: dict[str, list[tuple[str, str | None, Result]]] = {}

    def run(self) -> Result:
        """
        Run all processes for this stage
        """
        try:
            self.plugin_ref_list = self._get_stage_plugins()
            self.required_coverage = self._get_required_coverage()
        except (ModuleNotFoundError, HopprPluginError) as err:
            return Result.fail(str(err))

        result = self._check_bom_access()
        if not result.is_success():
            return result

        # Run each sub-stage (pre_stage_process, process_component, post_stage_process)
        # for all plugins (and, for process_component, for all components).
        # Each sub-stage must complete before the next can begin
        result.merge(self._run_all(HopprPlugin.pre_stage_process.__name__))
        result.merge(self._run_all(HopprPlugin.process_component.__name__))
        result.merge(self._run_all(HopprPlugin.post_stage_process.__name__))

        return result

    def _run_all(self, method_name: str):
        """
        Run the named method for all plugins.  If appropriate to the method, run it for
        all components for all plug-ins.
        """
        # Map to allow access to the arguments that went into a future call
        future_argument_map: dict[Future[Result], tuple[Plugin, Component | None]] = {}

        with ProcessPoolExecutor(max_workers=self.context.max_processes) as executor:
            for plugin in self.plugin_ref_list:
                if method_name in self.component_based_methods:
                    # Create one concurrent future object to run this method for each component
                    for component in self.context.delivered_sbom.components or []:
                        future_proc = executor.submit(
                            _run_plugin,
                            plugin_name=plugin.name,
                            context=self.context,
                            config=plugin.config,
                            method_name=method_name,
                            component=component,
                        )
                        future_argument_map[future_proc] = (plugin, component)
                else:
                    # Create a concurrent future object to run this method
                    future_proc = executor.submit(
                        _run_plugin,
                        plugin_name=plugin.name,
                        context=self.context,
                        config=plugin.config,
                        method_name=method_name,
                        component=None,
                    )
                    future_argument_map[future_proc] = (plugin, None)

            # Save all the results, count failures and retries
            # Note: future.results() blocks until the process is complete
            need_method_label = True
            failures = 0
            retries = 0

            for future_proc, (plugin, comp) in future_argument_map.items():
                future_result: Result = future_proc.result()
                if not future_result.is_skip():
                    if need_method_label:
                        echo(f"   Beginning method {method_name}")
                        need_method_label = False

                    plugin_cls = plugin_class(plugin.name)

                    self._save_result(method_name, plugin_cls.__name__, future_result, comp)
                    self._report_result(plugin_cls.__name__, comp, future_result)
                    self._update_bom(future_result.return_obj, comp)
                    ReportGenerator.report_gen_list.append(
                        Report(
                            uuid.uuid4(),
                            plugin_cls.__name__,
                            self.stage_id,
                            future_result.status,
                            future_result.message,
                            method_name,
                            comp,
                        )
                    )

                    if future_result.is_fail():
                        failures += 1
                    if future_result.is_retry():
                        retries += 1

        if method_name in self.component_based_methods:
            failures += self._check_component_coverage(method_name)

        return self._get_stage_result(method_name, failures, retries)

    def _get_stage_result(self, method_name: str, failures: int, retries: int) -> Result:
        if failures + retries == 0:
            return Result.success()
        if failures == 0 and retries > 0:
            return Result.fail(f"{retries} '{method_name}' processes returned 'retry'")
        if failures > 0 and retries == 0:
            return Result.fail(f"{failures} '{method_name}' processes failed")
        return Result.fail(f"{failures} '{method_name}' processes failed, and {retries} returned 'retry'")

    def _get_required_coverage(self) -> ComponentCoverage:
        if self.config_component_coverage is not None:
            return self.config_component_coverage

        if len(self.plugin_ref_list) == 0:
            return ComponentCoverage.OPTIONAL

        plugin = plugin_class(self.plugin_ref_list[0].name)
        coverage = plugin.default_component_coverage

        for plugin_ref in self.plugin_ref_list:
            plugin = plugin_class(plugin_ref.name)
            if plugin.default_component_coverage != coverage:
                raise HopprPluginError(
                    f"Plugins for stage {self.stage_id} do not have consistent default "
                    "component coverage values. The value may be overridden in transfer file."
                )

        return coverage

    def _check_component_coverage(self, method_name: str) -> int:
        result_count: dict[str | None, int] = {}
        for (_, purl, _) in self.results.get(method_name, []):
            result_count[purl] = result_count.get(purl, 0) + 1

        additional_failures = 0
        for component in self.context.delivered_sbom.components or []:
            count = result_count.get(component.purl, 0)
            if not self.required_coverage.accepts_count(count):
                bad_comp_result = Result.fail(
                    f"Component processed {count} times, {self.required_coverage.name} coverage required"
                )
                self._save_result(method_name, f"Stage {self.stage_id}", bad_comp_result, component)
                self._report_result(f"Stage {self.stage_id}", component, bad_comp_result)
                additional_failures += 1

        return additional_failures

    def _check_bom_access(self) -> Result:
        access_counts: dict[BomAccess, list[str]] = {}
        for access in BomAccess:
            access_counts[access] = []

        for plugin_ref in self.plugin_ref_list:
            plugin_cls = plugin_class(plugin_ref.name)
            access_counts[plugin_cls.bom_access].append(plugin_cls.__name__)

        if len(access_counts[BomAccess.FULL_ACCESS]) > 0 and len(self.plugin_ref_list) > 1:
            msg = (
                f"Stage {self.stage_id} has one or more plugins with {BomAccess.FULL_ACCESS.name}: "
                f"{', '.join(access_counts[BomAccess.FULL_ACCESS])}"
                ", and multiple plugins defined for the stage."
                "\n    Any plugin with FULL BOM access must be the only plugin in the stage"
            )
            secho(msg, fg=colors.RED)
            return Result.fail(msg)

        if len(access_counts[BomAccess.COMPONENT_ACCESS]) > 0 and self.required_coverage.max_value > 1:
            msg = (
                f"Stage {self.stage_id} has one or more plugins with {BomAccess.COMPONENT_ACCESS.name}: "
                f"{', '.join(access_counts[BomAccess.COMPONENT_ACCESS])}"
                f", and required component coverage for the stage of {self.required_coverage.name}."
                "\n    If any plugins have COMPONENT access, the stage required coverage must be "
                "EXACTLY_ONCE or NO_MORE_THAN_ONCE."
            )
            secho(msg, fg=colors.RED)
            return Result.fail(msg)

        return Result.success()

    def _update_bom(self, return_obj: Any, comp: Component | None):
        if self.context.delivered_sbom.components is not None and isinstance(return_obj, Component):
            for index, delivered_comp in enumerate(self.context.delivered_sbom.components):
                if delivered_comp == comp:
                    self.context.delivered_sbom.components[index] = return_obj
                    break

        elif isinstance(return_obj, Sbom):
            self.context.delivered_sbom = return_obj

    @staticmethod
    def _report_result(plugin: str, comp: Component | None, result: Result):
        desc = f"      {plugin} {result.status.name}"
        if comp is not None:
            desc = desc + f" for {comp.purl}"

        if result.is_success():
            color = colors.GREEN
        elif result.is_excluded():
            color = colors.YELLOW
        else:
            color = colors.RED
        if result.message:
            desc = desc + f": {result.message}"
        secho(desc, fg=color)

    def _save_result(self, method_name: str, plugin: str, result: Result, comp: Component | None):
        """
        Store the results for later use
        """
        comp_string = None
        if comp is not None and PackageURL.from_string(comp.purl) is not None:
            comp_string = comp.purl

        # If needed, create a new list for this method
        # Might need to expand this definition in the future to separate by plug-in

        if not method_name in self.results:
            self.results[method_name] = []

        self.results[method_name].append((plugin, comp_string, result))

    def _get_stage_plugins(self) -> list[Plugin]:
        """
        Determine list of plugin references used in this stage
        """
        used_purl_types: set[str] = set()

        # Get list of all component PURLs that are not None
        results = jmespath.search(
            expression="components[*].purl | not_null(@)",
            data=self.context.delivered_sbom.dict(by_alias=True),
        )

        used_purl_types.update(PackageURL.from_string(purl).type for purl in results)

        plugin_list: list[Plugin] = []

        # Loop over copy of transfer plugin references
        for plugin_ref in set(self.plugin_ref_list):
            plugin_cls = plugin_class(plugin_ref.name)

            # Determine if plugin's `supported_purl_types` are in the set of PURL types defined in SBOM components
            plugin_needed: bool = len(used_purl_types.intersection(plugin_cls.supported_purl_types)) > 0

            if len(plugin_cls.supported_purl_types) == 0 or plugin_needed:
                plugin_list.append(plugin_ref)

        return plugin_list


class HopprProcessor:  # pylint: disable=too-many-instance-attributes
    """
    Run the Hoppr process
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        transfer_file: Path,
        manifest_file: Path,
        credentials_file: Path | None,
        create_attestations: bool = False,
        functionary_key_path: Path | None = None,
        functionary_key_password: str | None = None,
        log_level: int = logging.INFO,
        log_file: str | Path | None = None,
        strict_repos: bool = True,
        previous_delivery: Path | None = None,
    ) -> None:
        self.context: HopprContext
        self.logger: MemoryLogger
        self.log_file: Path = Path(log_file or f"hoppr_{time.strftime('%Y%m%d-%H%M%S')}.log")
        self.log_level = log_level
        self.logfile_lock = multiprocessing.RLock()
        self.metadata_files: list[Path] = []
        self.strict_repos: bool = strict_repos

        self.manifest_file = ManifestFile.parse_file(manifest_file)

        self.credentials = Credentials.load(credentials_file)
        self.manifest = Manifest.load(manifest_file)
        self.transfer = Transfer.load(transfer_file)

        self.stage_processor_map: MutableMapping[StageRef, StageProcessor] = {}

        self.previous_delivery: Path | None = previous_delivery

        self.in_toto_links = HopprInTotoLinks(
            create_attestations,
            self.transfer,
            functionary_key_path,
            functionary_key_password,
        )

        self.report_gen: ReportGenerator

    def _collect_file(
        self, file_name: str | PathLike[str], target_dir: str | PathLike[str]
    ) -> None:  # pragma: no cover
        self.get_logger().info(f"Collecting metadata file {file_name}")
        abs_path = Path(file_name).absolute()

        target = Path(target_dir, quote_plus(f"{abs_path}"))
        shutil.copyfile(file_name, f"{target}")

    def _collect_manifest_metadata(self, manifest: ManifestFile, target_dir: str | PathLike[str]) -> None:
        for include_ref in manifest.includes:
            include = Manifest.loaded_manifests[include_ref]
            include_dict = include_ref.dict()

            if "local" in include_dict:
                url = str(include_dict["local"])
            else:
                url = str(include_dict["url"])

            target_file = Path(target_dir) / quote_plus(url)

            with target_file.open(mode="w+", encoding="utf-8") as output_file:
                output_file.write(include.yaml(by_alias=True, exclude_unset=True, indent=True))

        for sbom_ref in manifest.sboms:
            sbom = Sbom.loaded_sboms[sbom_ref]
            sbom_dict = sbom_ref.dict()

            if "local" in sbom_dict:
                url = str(sbom_dict["local"])
            elif "oci" in sbom_dict:
                url = str(sbom_dict["oci"])
            else:
                url = str(sbom_dict["url"])

            target_file = Path(target_dir) / quote_plus(url)

            with target_file.open(mode="w+", encoding="utf-8") as output_file:
                output_file.write(sbom.json(by_alias=True, exclude_unset=True, indent=2))

    def _collect_metadata(self):
        echo("Collecting Hoppr Metadata")
        target_dir = Path(self.context.collect_root_dir) / "generic" / "_metadata_"
        target_dir.mkdir(exist_ok=True, parents=True)

        if os.name == "posix":
            user = pwd.getpwuid(os.getuid()).pw_name
        else:
            user = os.getlogin()  # pragma: no cover

        with (target_dir / "_run_data_").open(mode="w", encoding="utf-8") as rundata:
            rundata.write(
                f"Hoppr Version:     {__version__}\n"
                f"Collection Start:  {str(datetime.now())}\n"
                f"User:              {user}\n"
                f"Host FQDN:         {socket.getfqdn()}\n"
                f"Working directory: {os.getcwd()}\n"
                f"Argument List:     {sys.argv[1:]}\n"
            )

        for file_name in self.metadata_files:
            self._collect_file(file_name, target_dir)

        self._collect_manifest_metadata(self.manifest_file, target_dir)

    def _collect_consolidated_bom(self) -> None:
        target_dir = Path(self.context.collect_root_dir) / "generic" / "_metadata_"
        target_dir.mkdir(parents=True, exist_ok=True)

        with (target_dir / "_consolidated_bom.json").open(mode="w", encoding="utf-8") as bom_data:
            bom_data.write(self.context.consolidated_sbom.json(exclude_none=True, by_alias=True, indent=2))

    def _collect_delivered_bom(self, stage_name: str | None = None) -> None:
        target_dir = Path(self.context.collect_root_dir) / "generic" / "_metadata_"
        target_dir.mkdir(parents=True, exist_ok=True)

        preface = f"intermediate_{stage_name}_" if stage_name else ""

        with (target_dir / f"_{preface}delivered_bom.json").open(mode="w", encoding="utf-8") as bom_data:
            bom_data.write(self.context.delivered_sbom.json(exclude_none=True, by_alias=True, indent=2))

    def _summarize_results(self) -> int:
        """
        Summarize the results of a HopprProcess run
        """
        echo(f"\n{' Results Summary '.center(40, '=')}")

        total_success_count = 0
        total_failure_count = 0
        for stage_ref, stage in self.stage_processor_map.items():
            echo(f"\nStage: {stage_ref.name}")

            for method_name, result_list in stage.results.items():
                echo(f"   {method_name}")
                result_count = len(result_list)
                failure_count = 0
                failure_list = "\n      Failure Summary:\n"

                for plugin_name, comp_str, result in result_list:
                    # All retries should be handled internally by the plugins,
                    # So if a RETRY result is returned, that's a failure
                    if result.is_fail() or result.is_retry():
                        failure_count += 1
                        failure_list += f"         {plugin_name}: "

                        if comp_str is not None:
                            failure_list += f"Component: {comp_str}: "

                        failure_list += result.message + "\n"

                total_failure_count += failure_count
                total_success_count += result_count - failure_count
                echo(f"      {result_count - failure_count} jobs succeeded, {failure_count} failed")

                if failure_count > 0:
                    echo(failure_list)

        echo(f"\nGRAND TOTAL: {total_success_count} jobs succeeded, {total_failure_count} failed\n")
        self.report_gen = ReportGenerator(self.context)
        self.report_gen.generate_report()

        return total_failure_count

    def get_logger(
        self,
        lock: RLock | None = None,
        log_name: str | None = None,
        log_file: str | Path = "hoppr.log",
        log_level: int = logging.INFO,
    ) -> MemoryLogger:
        """
        Returns the logger for this class

        Args:
            lock (RLock | None, optional): Reentrant lock for shared log file. Defaults to None.
            log_name (str | None, optional): Name of logger. Defaults to None.
            log_file (str | Path | None, optional): Path to log file. Defaults to "hoppr.log".
            log_level (int, optional): Logging level. Defaults to logging.INFO.

        Returns:
            MemoryLogger: A new MemoryLogger instance.
        """
        if not hasattr(self, "logger"):
            self.logger = MemoryLogger(
                file_name=str(log_file),
                lock=lock,
                log_name=log_name or f"HopprProcessor--{os.getpid()}",
                log_level=log_level,
                flush_immed=True,
            )

        return self.logger

    def run(self) -> Result:
        """
        Run the Hoppr process executing each stage in turn

        Args:
            log_file (Path | None, optional): Path to log file. Defaults to None.
            strict_repos (bool, optional): Enables strict repository searches for components. Defaults to True.

        Returns:
            Result: The composite result of all executed stages
        """
        result = Result.success()

        with tempfile.TemporaryDirectory() as collection_root, multiprocessing.Manager() as manager:
            logfile_lock = manager.RLock()

            self.context = HopprContext(
                collect_root_dir=Path(collection_root).resolve(),
                consolidated_sbom=deepcopy(Sbom.consolidated_sbom),
                credential_required_services=getattr(self.credentials, "credential_required_services", None),
                delivered_sbom=deepcopy(Sbom.consolidated_sbom),
                log_level=self.log_level,
                logfile_location=self.log_file,
                logfile_lock=logfile_lock,
                max_processes=self.transfer.max_processes or cpu_count(),
                repositories=self.manifest.repositories,
                sboms=list(Sbom.loaded_sboms.values()),
                stages=self.transfer.stages,
                strict_repos=self.strict_repos,
            )

            self.logger = self.get_logger(
                lock=logfile_lock,
                log_name=f"HopprProcessor--{os.getpid()}",
                log_file=self.log_file,
                log_level=self.log_level,
            )

            if self.context.consolidated_sbom is not None and len(self.context.consolidated_sbom.components or []) == 0:
                msg = "No SBOMs defined in manifests, or SBOMs contain no components. Nothing to process."
                secho(msg, fg=colors.RED)
                return Result.fail(msg)

            self.in_toto_links.set_collection_root(collection_root)
            self.in_toto_links.record_stage_start("_collect_metadata")
            self._collect_metadata()
            self._collect_consolidated_bom()
            self.in_toto_links.record_stage_stop("_collect_metadata")

            # Reset some class variables (we should refactor to use instance variables, see issue
            # https://gitlab.com/hoppr/hoppr/-/issues/230) so that:
            #   1: The consolidated bom cannot be modified
            #   2: Plug-ins may choose to load other manifests.
            del Sbom.consolidated_sbom
            Component.component_lookup = {}
            Component.components = []
            Component.properties = []
            Manifest.loaded_manifests = {}

            msg = f"Beginning Hoppr Process execution, max_processes={self.context.max_processes}"
            self.logger.info(msg=msg)
            echo(message=msg)

            for stage_ref in self.transfer.stages:
                msg = f" Beginning Stage {stage_ref.name} ".center(100, "=")
                self.logger.info(msg)
                echo(msg)

                stage = StageProcessor(stage_ref, self.context)
                self.in_toto_links.record_stage_start(stage_ref.name)
                self.stage_processor_map[stage_ref] = stage

                result = stage.run()

                if result.is_fail() or result.is_retry():
                    msg = f"Stage {stage_ref.name} failed, processing terminated: {result.message}"
                    self.logger.error(msg=msg)
                    secho(message=f"  {msg}", fg=colors.RED)

                    break

                self._collect_delivered_bom(stage_ref.name)
                self._collect_delivered_bom()
                self.in_toto_links.record_stage_stop(stage_ref.name)

            self.in_toto_links.record_stage_start("_finalize")
            self._collect_delivered_bom()
            self.in_toto_links.record_stage_stop("_finalize")

        failed_jobs = self._summarize_results()
        if failed_jobs > 0:
            result.merge(Result.fail(f"{failed_jobs} failed during this execution"))

        return result



"""
Define constants used by Hoppr
"""

from enum import Enum


class BomProps(str, Enum):
    """
    SBOM Property Field Names
    """

    COLLECTION_DIRECTORY = "hoppr:collection:directory"
    COLLECTION_PLUGIN = "hoppr:collection:plugin"
    COLLECTION_REPOSITORY = "hoppr:collection:repository"
    COLLECTION_TIMETAG = "hoppr:collection:timetag"

    COMPONENT_SEARCH_SEQUENCE = "hoppr:repository:component_search_sequence"


class ConfigKeys(str, Enum):
    """
    Keys used in configuration files
    """

    CONFIG = "config"
    PLUGINS = "plugins"



"""
Collector plugin for pypi images
"""
from __future__ import annotations

import importlib.util
import re
import sys

from subprocess import CalledProcessError

from packageurl import PackageURL

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models import HopprContext
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.manifest import Component
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result


class CollectPypiPlugin(SerialCollectorPlugin):
    """
    Collector plugin for pypi images
    """

    supported_purl_types = ["pypi"]
    required_commands = [sys.executable]
    products: list[str] = ["pypi/*"]
    system_repositories = ["https://pypi.org/simple"]

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)

        self.manifest_repos: list[str] = []
        self.password_list: list[str] = []
        self.base_command = [self.required_commands[0], "-m", "pip"]

    def _run_cmd_wrapper(  # pylint: disable=too-many-arguments
        self, command, password_list, param: str, pkg_format: str, log_if_error: str
    ) -> bool:
        """
        Run command utility for discrete function calls - required for DO-178C Level-A branch isolation coverage
        """
        run_result = self.run_command([*command, param], password_list)
        try:
            run_result.check_returncode()
            return True
        except CalledProcessError:
            self.get_logger().debug(msg=f"{log_if_error} {pkg_format} package", indent_level=2)
        return False

    def collect_binary_only(self, command, password_list, log_if_error: str) -> bool:
        """
        Only collect the PyPI binary (WHL)
        """
        return self._run_cmd_wrapper(command, password_list, "--only-binary=:all:", "binary", log_if_error)

    def collect_source(self, command, password_list, log_if_error: str) -> bool:
        """
        Only collect the PyPI source package
        """
        return self._run_cmd_wrapper(command, password_list, "--no-binary=:all:", "source", log_if_error)

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        """
        Copy a component to the local collection directory structure
        """
        if importlib.util.find_spec(name="pip") is None:
            return Result.fail(message="The pip package was not found. Please install and try again.")

        purl = PackageURL.from_string(comp.purl)

        source_url = RepositoryUrl(url=repo_url)
        if not re.match(pattern="^.*simple/?$", string=f"{source_url}"):
            source_url /= "simple"

        password_list = []

        if creds is not None:
            source_url = RepositoryUrl(
                url=source_url.url,
                username=creds.username,
                password=creds.password.get_secret_value(),
            )
            password_list = [creds.password.get_secret_value()]

        target_dir = self.directory_for(purl.type, repo_url, subdir=f"{purl.name}_{purl.version}")

        self.get_logger().info(msg=f"Target directory: {target_dir}", indent_level=2)

        command = [
            *self.base_command,
            "download",
            "--no-deps",
            "--no-cache",
            "--timeout",
            "60",
            "--index-url",
            f"{source_url}",
            "--dest",
            f"{target_dir}",
            f"{purl.name}=={purl.version}",
        ]

        base_error_msg = f"Failed to download {purl.name} version {purl.version}"

        if self.collect_binary_only(command, password_list, base_error_msg):  # pylint: disable=no-else-return
            self.set_collection_params(comp, repo_url, target_dir)
            return Result.success(return_obj=comp)
        elif self.collect_source(command, password_list, base_error_msg):  # pylint: disable=no-else-return
            self.set_collection_params(comp, repo_url, target_dir)
            return Result.success(return_obj=comp)
        else:
            return Result.retry(f"{base_error_msg} as either binary or source.")



"""
A logger that gets dumped to stdout when closed
"""
from __future__ import annotations

import inspect
import itertools
import logging
import os

from logging.handlers import MemoryHandler
from threading import _RLock as RLock


class MemoryLogger:
    """
    A logger that gets dumped to stdout when closed
    """

    id_iter = itertools.count()

    def __init__(  # pylint: disable=too-many-arguments
        self,
        file_name: str,
        lock: RLock | None,
        log_name: str | None = None,
        log_level: int = logging.INFO,
        flush_immed: bool = False,
    ) -> None:
        self.instance_id = next(self.id_iter)
        self.flush_immed = flush_immed
        self.lock: RLock | None = lock

        if log_name is None:
            caller = inspect.stack()[1].function

            caller_cls = inspect.stack()[1].frame.f_locals.get("self", None)
            if caller_cls is not None:
                caller = type(caller_cls).__name__

            log_name = f"{caller}--{os.getpid()}-{self.instance_id}"

        self._logger = logging.getLogger(log_name)
        self._logger.setLevel(log_level)

        formatter = logging.Formatter(fmt="[$asctime] - [$name] - [$levelname] - $message", style="$")

        file_handler = logging.FileHandler(file_name)
        file_handler.setFormatter(formatter)

        log_handler = MemoryHandler(10000, flushLevel=logging.CRITICAL, target=file_handler)
        log_handler.setFormatter(formatter)

        self._logger.addHandler(log_handler)

    def flush(self) -> None:
        """
        Flush all handlers for this logger
        """
        if self.lock is not None:
            with self.lock:
                for handler in self._logger.handlers:
                    handler.flush()
        else:
            for handler in self._logger.handlers:
                handler.flush()

    def close(self) -> None:
        """
        Close (and flush) all handlers for this logger
        """
        if self.lock is not None:
            with self.lock:
                for handler in self._logger.handlers:
                    handler.close()
        else:
            for handler in self._logger.handlers:
                handler.close()

    def clear_targets(self) -> None:
        """
        Makes the target for all Memory Handlers in this logger None

        Thus when these handlers are flushed, nothing will go to standard output
        """
        for handler in self._logger.handlers:
            if isinstance(handler, MemoryHandler):
                handler.setTarget(None)

    def log(self, level: int, msg: str, indent_level: int) -> None:
        """
        Wrapper function for logging messages
        """
        indent_string = " " * 4 * indent_level

        if self.lock is not None:
            self.lock.acquire()

        self._logger.log(level, indent_string + msg)
        if self.flush_immed:
            self.flush()

        if self.lock is not None:
            self.lock.release()

    def debug(self, msg: str, indent_level: int = 0) -> None:
        """
        Wrapper function for debug messages
        """
        self.log(logging.DEBUG, msg, indent_level)

    def info(self, msg: str, indent_level: int = 0) -> None:
        """
        Wrapper function for info messages
        """
        self.log(logging.INFO, msg, indent_level)

    def warning(self, msg: str, indent_level: int = 0) -> None:
        """
        Wrapper function for warn messages
        """
        self.log(logging.WARNING, msg, indent_level)

    def error(self, msg: str, indent_level: int = 0) -> None:
        """
        Wrapper function for error messages
        """
        self.log(logging.ERROR, msg, indent_level)

    def fatal(self, msg: str, indent_level: int = 0) -> None:
        """
        Wrapper function for fatal messages
        """
        self.log(logging.FATAL, msg, indent_level)

    def critical(self, msg: str, indent_level: int = 0) -> None:
        """
        Wrapper function for critical messages
        """
        self.log(logging.CRITICAL, msg, indent_level)



"""
Framework for manipulating bundles for airgapped transfers.
"""
import logging
import sys

from pathlib import Path
from typing import List, Optional

from typer import Exit, echo, prompt

from hoppr import __version__
from hoppr.in_toto import generate_in_toto_layout
from hoppr.models.credentials import Credentials
from hoppr.models.manifest import Manifest
from hoppr.models.transfer import Transfer
from hoppr.processor import HopprProcessor


def bundle(  # pylint: disable=too-many-arguments,too-many-locals
    manifest_file: Path,
    credentials_file: Path,
    transfer_file: Path,
    log_file: Path,
    verbose: bool = False,
    strict_repos: bool = True,
    create_attestations: bool = False,
    functionary_key_path: Optional[Path] = None,
    functionary_key_prompt: bool = False,
    functionary_key_password: Optional[str] = None,
    previous_delivery: Optional[Path] = None,
):
    """
    Run the stages specified in the transfer config
    file on the content specified in the manifest
    """
    metadata_files = [manifest_file, transfer_file]

    if credentials_file is not None:
        metadata_files.append(credentials_file)

    log_level = logging.DEBUG if verbose else logging.INFO

    if create_attestations and functionary_key_path is None:
        echo("To create attestations both the `--attest` option and a functionary private key need to be provided.")
        raise Exit(code=1)

    if functionary_key_prompt:
        functionary_key_password = prompt(f"Enter password for {str(functionary_key_path)}", hide_input=True)

    processor = HopprProcessor(
        transfer_file=transfer_file,
        manifest_file=manifest_file,
        credentials_file=credentials_file,
        create_attestations=create_attestations,
        functionary_key_path=functionary_key_path,
        functionary_key_password=functionary_key_password,
        log_level=log_level,
        log_file=log_file,
        strict_repos=strict_repos,
        previous_delivery=previous_delivery,
    )

    processor.metadata_files = metadata_files

    result = processor.run()

    if result.is_fail():
        sys.exit(1)


def generate_layout(
    transfer_file: Path,
    project_owner_key_path: Path,
    functionary_key_path: Path,
    project_owner_key_prompt: bool,
    project_owner_key_password: str,
):
    """
    Create in-toto layout based on transfer file.
    """

    if project_owner_key_prompt:
        project_owner_key_password = prompt(f"Enter password for {str(project_owner_key_path)}", hide_input=True)

    transfer = Transfer.load(transfer_file)
    generate_in_toto_layout(transfer, project_owner_key_path, functionary_key_path, project_owner_key_password)


def validate(input_files: List[Path], credentials_file: Path, transfer_file: Path):
    """
    Validate multiple manifest files for schema errors.
    """

    if credentials_file is not None:
        Credentials.load(credentials_file)

    if transfer_file is not None:
        Transfer.load(transfer_file)

    for file in input_files:
        Manifest.load(file)



"""
Supporting class for oras bundle plugin
"""
import json
import os
import platform

from datetime import datetime
from http import cookiejar

import jsonschema  # type: ignore
import oras.defaults  # type: ignore
import oras.oci  # type: ignore
import oras.provider  # type: ignore
import requests

from oras.container import Container  # type: ignore
from oras.decorator import ensure_container  # type: ignore

import hoppr


class Registry(oras.provider.Registry):
    """Override the default Oras Registry Class"""

    @ensure_container
    def push(self, container: Container, archives: list, logger):  # pylint: disable=W0221
        """
        Given a list of layer metadata (paths and corresponding mediaType) push.
        """
        # Prepare a new manifest
        manifest = oras.oci.NewManifest()
        self.session.cookies.set_policy(BlockAll())

        self.upload_blobs(container, archives, manifest, logger)

        # Prepare manifest and config (add your custom annotations here)
        manifest["annotations"] = {
            "org.opencontainers.image.created": str(datetime.now()),
            "ArtifactType": "Oras OCI Bundle",
            "Documentation": "https://oras.land/cli/",
        }
        config_file = "/tmp/oras-conf.json"
        config_blob = {'os': str(platform.platform()), 'HopctlVersion': hoppr.__version__}
        self.setup_default_config(config_blob, config_file)
        conf, config_file = oras.oci.ManifestConfig(
            path=config_file, media_type="application/vnd.oci.image.config.v1+json"
        )
        conf["annotations"] = config_blob

        # Config is just another layer blob!
        logger.info(f"Uploading config to {container.uri}")
        logger.info(f"Config {conf}")
        logger.flush()
        response = self.upload_blob(config_file, container, conf)
        self._check_200_response(response)
        os.remove(config_file)

        # Final upload of the manifest
        manifest["config"] = conf
        # Try updating the manifest
        response = self.upload_manifest(manifest=manifest, container=container)
        self._check_200_response(response=response)
        return response

    def upload_blobs(self, container: Container, archives: list, manifest: dict, logger) -> dict:
        """
        Upload individual layers to oci registry

        Args:
            container (Container): oras container
            archives (list): list of layers to upload
            logger (_type_): hoppr logger
        """
        # Upload files as blobs
        for item in archives:

            blob = item.get("path")
            media_type = item.get("media_type")
            annots = item.get("annotations") or {}

            if not blob or not os.path.exists(blob):
                logger.info(f"Path {blob} does not exist or is not defined.")
                continue

            # Artifact title is basename or user defined
            blob_name = item.get("title") or os.path.basename(blob)

            # Create a new layer from the blob
            layer = oras.oci.NewLayer(blob, media_type, is_dir=False)
            logger.info(f"Preparing layer {layer}")

            # Update annotations with title we will need for extraction
            annots.update({oras.defaults.annotation_title: blob_name})
            layer["annotations"] = annots

            # update the manifest with the new layer
            manifest["layers"].append(layer)

            # Upload the blob layer
            logger.info(f"Uploading {blob} to {container.uri}")
            logger.flush()
            response = self.upload_blob(blob, container, layer)
            self._check_200_response(response)
        return manifest

    def upload_manifest(self, manifest: dict, container: Container) -> requests.Response:
        """
        Read a manifest file and upload it.

        :param manifest: manifest to upload
        :type manifest: dict
        :param container:  parsed container URI
        :type container: oras.container.Container or str
        """
        self.reset_basic_auth()  # Should we really be reseting?
        jsonschema.validate(manifest, schema=oras.schemas.manifest)
        headers = {
            "Content-Type": oras.defaults.default_manifest_media_type,
            "Content-Length": str(len(manifest)),
        }
        hostname = ""
        if ':' in container.registry:
            hostname = container.registry.split(":")[0]
        else:
            hostname = container.registry
        put_url = f"{self.prefix}://{hostname}/v2/{container.api_prefix}/manifests/{container.tag}"
        response = self.do_request(put_url, "PUT", headers=headers, json=manifest)
        return response

    def setup_default_config(self, config_blob: dict, config_file: str):
        """
        Setup defualt oras object config
        """
        with open(config_file, "w", encoding='utf-8') as fout:
            json.dump(config_blob, fout, indent=4)


class BlockAll(cookiejar.CookiePolicy):
    """
    Block all cookies

    Args:
        cookiejar (CookiePolicy): _description_
    """

    return_ok = (
        set_ok
    ) = domain_return_ok = path_return_ok = lambda self, *args, **kwargs: False  # pylint: disable=C3001
    netscape = True
    rfc2965 = hide_cookie2 = False



"""
Framework for manipulating bundles for airgapped transfers.
"""
from __future__ import annotations

import ctypes
import os
import sys

from pathlib import Path
from platform import python_version
from runpy import run_module
from typing import Optional

from typer import Argument, Option, Typer, echo, secho, style

from hoppr import __version__, main

# Windows flags and types
NT_ENABLE_ECHO_INPUT = 0x0004
NT_ENABLE_LINE_INPUT = 0x0002
NT_ENABLE_PROCESSED_INPUT = 0x0001
NT_CONSOLE_FLAGS = NT_ENABLE_ECHO_INPUT | NT_ENABLE_LINE_INPUT | NT_ENABLE_PROCESSED_INPUT
NT_STD_OUTPUT_HANDLE = ctypes.c_uint(-11)

# Enable ANSI processing on Windows systems
if sys.platform == "win32":  # pragma: no cover
    nt_kernel = ctypes.WinDLL(name="kernel32.dll")

    nt_kernel.SetConsoleMode(nt_kernel.GetStdHandle(NT_STD_OUTPUT_HANDLE), NT_CONSOLE_FLAGS)


ATTESTATION_OPTION: bool = Option(
    False,
    "-a",
    "--attest",
    help="Generate in-toto attestations",
    envvar="HOPPR_ATTESTATION",
)


CREDENTIALS_FILE_OPTION: Path = Option(
    None,
    "-c",
    "--credentials",
    help="Specify credentials config for services",
    envvar="HOPPR_CREDS_CONFIG",
)


FUNCTIONARY_KEY_BUNDLE_OPTION: Optional[Path] = Option(
    None,
    "-fk",
    "--functionary-key",
    help="Path to key used to sign in-toto layout",
    envvar="HOPPR_FUNCTIONARY_KEY",
)


FUNCTIONARY_KEY_GENERATE_LAYOUT_OPTION: Path = Option(
    ...,
    "-fk",
    "--functionary-key",
    help="Path to key used to sign in-toto layout",
    envvar="HOPPR_FUNCTIONARY_KEY",
)


FUNCTIONARY_KEY_PASSWORD_OPTION: str = Option(
    None,
    "-fk-pass",
    "--project-owner-key-password",
    help="Password for project owner key",
    envvar="HOPPR_PROJECT_OWNER_KEY_PASSWORD",
)


FUNCTIONARY_KEY_PROMPT_OPTION: bool = Option(
    False,
    "-p",
    "--prompt",
    help="Prompt user for project owner key's password",
    envvar="HOPPR_PROJECT_OWNER_KEY_PROMPT",
)


LOG_FILE_OPTION: Path = Option(
    None,
    "-l",
    "--log",
    help="File to which log will be written",
    envvar="HOPPR_LOG_FILE",
)


MANIFEST_FILE_ARGUMENT: Path = Argument(
    "manifest.yml",
    help="Path to manifest file",
    expose_value=True,
)


OUTPUT_DIR_OPTION: Path = Option(
    Path(),
    "-o",
    "--output-dir",
    file_okay=False,
    resolve_path=True,
    exists=True,
    show_default=False,
    help="[default: current directory]",
)


PREVIOUS_DELIVERY_OPTION: Path = Option(
    None,
    "-pd",
    "--previous-delivery",
    help="Path to manifest or tar bundle representing a previous delivery",
    envvar="HOPPR_PREVIOUS_DELIVERY",
)


PROJECT_OWNER_KEY_OPTION: Path = Option(
    ...,
    "-pk",
    "--project-owner-key",
    help="Path to key used to sign in-toto layout",
    envvar="HOPPR_PROJECT_OWNER_KEY",
)


PROJECT_OWNER_KEY_PASSWORD_OPTION: str = Option(
    None,
    "-pk-pass",
    "--project-owner-key-password",
    help="Password for project owner key",
    envvar="HOPPR_PROJECT_OWNER_KEY_PASSWORD",
)


PROJECT_OWNER_KEY_PROMPT_OPTION: bool = Option(
    False,
    "-p",
    "--prompt",
    help="Prompt user for project owner key's password",
    envvar="HOPPR_PROJECT_OWNER_KEY_PROMPT",
)


STRICT_OPTION: bool = Option(
    True,
    "--strict/--no-strict",
    help="Utilize only manifest repositories for package collection",
    envvar="HOPPR_STRICT_REPOS",
)


TRANSFER_FILE_OPTION: Path = Option(
    "transfer.yml",
    "-t",
    "--transfer",
    help="Specify transfer config",
    envvar="HOPPR_TRANSFER_CONFIG",
)


VERBOSE_OPTION: bool = Option(
    False,
    "-v",
    "--debug",
    "--verbose",
    help="Enable debug output",
)


app = Typer(
    context_settings=dict(help_option_names=["-h", "--help"]),
    help="Collect, process, & bundle your software supply chain",
    no_args_is_help=True,
)


@app.command()
def bundle(
    # pylint: disable=unused-argument
    # pylint: disable=too-many-arguments
    manifest_file: Path = MANIFEST_FILE_ARGUMENT,
    credentials_file: Path = CREDENTIALS_FILE_OPTION,
    transfer_file: Path = TRANSFER_FILE_OPTION,
    log_file: Path = LOG_FILE_OPTION,
    verbose: bool = VERBOSE_OPTION,
    strict_repos: bool = STRICT_OPTION,
    create_attestations: bool = ATTESTATION_OPTION,
    functionary_key_path: Optional[Path] = FUNCTIONARY_KEY_BUNDLE_OPTION,
    functionary_key_prompt: bool = FUNCTIONARY_KEY_PROMPT_OPTION,
    functionary_key_password: str = FUNCTIONARY_KEY_PASSWORD_OPTION,
    previous_delivery: Path = PREVIOUS_DELIVERY_OPTION,
):  # pragma: no cover
    """
    Run the stages specified in the transfer config
    file on the content specified in the manifest
    """
    main.bundle(**locals())


@app.command()
def validate(
    # pylint: disable=unused-argument
    input_files: list[Path],
    credentials_file: Path = CREDENTIALS_FILE_OPTION,
    transfer_file: Path = TRANSFER_FILE_OPTION,
):  # pragma: no cover
    """
    Validate multiple manifest files for schema errors
    """
    main.validate(**locals())


@app.command()
def generate_layout(
    # pylint: disable=unused-argument
    transfer_file: Path = TRANSFER_FILE_OPTION,
    project_owner_key_path: Path = PROJECT_OWNER_KEY_OPTION,
    functionary_key_path: Path = FUNCTIONARY_KEY_GENERATE_LAYOUT_OPTION,
    project_owner_key_prompt: bool = PROJECT_OWNER_KEY_PROMPT_OPTION,
    project_owner_key_password: str = PROJECT_OWNER_KEY_PASSWORD_OPTION,
):  # pragma: no cover
    """
    Create in-toto layout based on transfer file
    """
    main.generate_layout(**locals())


@app.command()
def generate_schemas(output_dir: Path = OUTPUT_DIR_OPTION):  # pragma: no cover
    """
    Generate schema files from input file models
    """
    os.chdir(output_dir)
    run_module(mod_name="hoppr.models")


@app.command()
def version():
    """
    Print version information for `hoppr`
    """
    # Only print ANSI art on TTY terminals
    if sys.stdout.isatty():
        hippo = Path(__file__).parent / ".." / "resources" / "hoppr-hippo.ansi"
        with hippo.open(mode="rb") as ansi:
            echo(ansi.read())

    secho(f"{style(text='Hoppr Framework Version', fg='green')} : {__version__}")
    secho(f"{style(text='Python Version         ', fg='green')} : {python_version()}")



"""
Collector plugin for git repositories
"""
from __future__ import annotations

import os
import re

from pathlib import Path

from packageurl import PackageURL  # type: ignore[import]

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models import HopprContext
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.manifest import Component
from hoppr.result import Result


class CollectGitPlugin(SerialCollectorPlugin):
    """
    Class to copy git repositories

    configuration options:
    config:
      depth: all, 1, or any int > 1

    """

    supported_purl_types = ["git", "gitlab", "github"]
    required_commands = ["git"]
    products: list[str] = ["git/*", "gitlab/*", "github/*"]

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)
        if self.config is not None and "git_command" in self.config:
            self.required_commands = [self.config["git_command"]]

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        """
        Collect git repository
        """
        purl: PackageURL = PackageURL.from_string(comp.purl)  # pyright: ignore[reportGeneralTypeIssues]

        source_url = os.path.join(repo_url, purl.namespace, purl.name)

        self.get_logger().info(msg=f"Cloning {source_url}", indent_level=2)

        # Only reference the namespace.  The actual clone creates the named directory
        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)

        git_result = self.git_clone(tmp_dir=target_dir, source_url=source_url, source_creds=creds, comp=comp)
        if not git_result.is_success():
            return git_result

        git_result = self.git_update(target_dir, purl.name)
        if not git_result.is_success():
            return git_result

        self.set_collection_params(comp, repo_url, target_dir)

        return Result.success(return_obj=comp)

    def git_clone(
        self,
        tmp_dir: Path,
        source_url: str,
        source_creds: CredentialRequiredService | None,
        comp: Component,
    ) -> Result:
        """
        Git clone
        """
        git_src = source_url
        password_list: list[str] = []
        if re.match("^https?://", git_src) and source_creds is not None:
            git_src = git_src.replace("://", f"://{source_creds.username}@", 1)
            if source_creds.password:
                git_src = git_src.replace("@", f":{source_creds.password.get_secret_value()}@")
                password_list = [source_creds.password.get_secret_value()]

        if git_src.startswith("ssh://") and source_creds is not None:
            git_src = git_src.replace("ssh://", f"ssh://{source_creds.username}@")

        if not git_src.endswith(".git"):
            git_src += ".git"

        opts: list[str] = []
        if str(comp.version).strip():
            opts.extend(("--branch", str(comp.version)))

        # Default depth
        depth = "1"
        if self.config is not None:
            # Allow for further depth default to 1
            depth = self.config.get("depth", depth)

        # Recognize the 'all' keyword as requesting all history
        if depth.lower() != "all":
            opts.extend(("--depth", depth))

        # Command
        command = [self.required_commands[0], "clone", *opts, git_src]

        # Only clone with a depth of one and reference the version specified.
        result = self.run_command(
            command,
            password_list,
            cwd=tmp_dir,
        )
        if result.returncode != 0:
            msg = f"Failed to clone {source_url}"
            self.get_logger().debug(msg=msg, indent_level=2)
            return Result.retry(message=msg)

        return Result.success()

    def git_update(self, tmp_dir: Path, name_git: str) -> Result:
        """
        Git update-server-info
        """
        repo_dir = tmp_dir / name_git

        # Make the clone usable as a remote
        result = self.run_command(
            [self.required_commands[0], "update-server-info"],
            cwd=repo_dir,
        )

        if result.returncode != 0:
            msg = "Failed to make the clone usable as a remote"
            self.get_logger().debug(msg=msg, indent_level=2)
            return Result.retry(message=msg)

        return Result.success()



"""
Collector plugin for Yum packages
"""
from __future__ import annotations

from pathlib import Path
from typing import Any

from packageurl import PackageURL

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models import HopprContext
from hoppr.result import Result


def _artifact_string(purl):
    artifact_string = purl.name
    if purl.version is not None:
        artifact_string += f"-{purl.version}"

    # Limiting to an architecture using --archlist finds all compatible architectures,
    # not just the specified one, so . . .
    if "arch" in purl.qualifiers:
        artifact_string += f"*{purl.qualifiers.get('arch')}"

    return artifact_string


class CollectYumPlugin(SerialCollectorPlugin):
    """
    Collector plugin for yum images
    """

    supported_purl_types = ["rpm"]
    required_commands = ["yumdownloader"]
    products: list[str] = ["rpm/*"]

    def get_version(self) -> str:
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)
        if self.config is not None:
            if "yumdownloader_command" in self.config:
                self.required_commands = [self.config["yumdownloader_command"]]

    @hoppr_rerunner
    def collect(self, comp: Any, repo_url: str, creds=None):
        """
        Copy a component to the local collection directory structure
        """

        purl: PackageURL = PackageURL.from_string(comp.purl)  # pyright: ignore[reportGeneralTypeIssues]
        artifact = _artifact_string(purl)

        self.get_logger().info(msg=f"Copying yum package from {comp.purl}", indent_level=2)

        command = [
            self.required_commands[0],
            "-q",
            "--disableexcludes=all",
            "--setopt=*.module_hotfixes=true",
            "--downloadonly",
            "--urls",
            artifact,
        ]

        run_result = self.run_command(command)
        if run_result.returncode != 0:
            msg = (
                f"{self.required_commands[0]} failed to locate package for {comp.purl}, "
                f"return_code={run_result.returncode}"
            )
            self.get_logger().debug(msg=msg, indent_level=2)

            return Result.retry(message=msg)

        # Taking the first URL if multiple are returned
        found_url = run_result.stdout.decode("utf-8").strip().split("\n")[0]

        if not found_url.startswith(repo_url):
            msg = "Yum download url does not match requested url."

            self.get_logger().error(msg=msg, indent_level=2)
            self.get_logger().error(msg=f"Yum download url:      {found_url}", indent_level=3)
            self.get_logger().error(msg=f"Expected url starting: {repo_url}", indent_level=3)

            return Result.fail(message=msg)

        found_url_path = Path(found_url)
        subdir = "/".join(found_url_path.parts[len(Path(repo_url).parts) : -1])
        target_dir = self.directory_for(purl.type, repo_url, subdir=subdir)

        command = [
            self.required_commands[0],
            "-q",
            "--disableexcludes=all",
            "--setopt=*.module_hotfixes=true",
            "--downloadonly",
            f"--destdir={target_dir}",
            artifact,
        ]

        run_result = self.run_command(command, [])
        if run_result.returncode != 0:
            msg = f"Failed to download Yum artifact {purl.name} version {purl.version} "
            return Result.retry(message=msg)

        self.set_collection_params(comp, repo_url, target_dir)

        return Result.success(return_obj=comp)
